"""Functions for subsampling datasets.

"""
import warnings

import pandas as pd
from pymbar.timeseries import detect_equilibration as _detect_equilibration
from pymbar.timeseries import statistical_inefficiency as _statistical_inefficiency
from pymbar.timeseries import subsample_correlated_data as _subsample_correlated_data

from .. import pass_attrs


def decorrelate_u_nk(
    df, method="dE", drop_duplicates=True, sort=True, remove_burnin=False, **kwargs
):
    """Subsample an u_nk DataFrame based on the selected method.

    The method can be either 'all' (obtained as a sum over all energy
    components) or 'dE'. In the latter case the energy differences
    :math:`dE_{i,i+1}` (:math:`dE_{i,i-1}` for the last lambda) are used.  This
    is a wrapper function around the function
    :func:`~alchemlyb.preprocessing.subsampling.statistical_inefficiency` or
    :func:`~alchemlyb.preprocessing.subsampling.equilibrium_detection`.

    Parameters
    ----------
    df : DataFrame
        DataFrame to be subsampled according to the selected method.
    method : {'all', 'dE'}
        Method for decorrelating the data.
    drop_duplicates : bool
        Drop the duplicated lines based on time.
    sort : bool
        Sort the Dataframe based on the time column.
    remove_burnin : bool
        Whether to perform equilibrium detection (``True``) or just do
        statistical inefficiency (``False``).

        .. versionadded:: 1.0.0

    **kwargs :
        Additional keyword arguments for
        :func:`~alchemlyb.preprocessing.subsampling.statistical_inefficiency`
        or :func:`~alchemlyb.preprocessing.subsampling.equilibrium_detection`.
    Returns
    -------
    DataFrame
        `df` subsampled according to selected `method`.
    Note
    ----
    The default of ``True`` for  `drop_duplicates` and `sort` should result in robust decorrelation
    but can lose data.


    .. versionadded:: 0.6.0
    .. versionchanged:: 1.0.0
       Add the `remove_burnin` keyword to allow unequilibrated frames
       to be removed. Rename `method` value 'dhdl_all' to 'all' and
       deprecate the 'dhdl'.

    """
    kwargs["drop_duplicates"] = drop_duplicates
    kwargs["sort"] = sort

    series = u_nk2series(df, method)

    if remove_burnin:
        return equilibrium_detection(df, series, **kwargs)
    else:
        return statistical_inefficiency(df, series, **kwargs)


def decorrelate_dhdl(
    df, drop_duplicates=True, sort=True, remove_burnin=False, **kwargs
):
    """Subsample a dhdl DataFrame.
    This is a wrapper function around the function
    :func:`~alchemlyb.preprocessing.subsampling.statistical_inefficiency` and
    :func:`~alchemlyb.preprocessing.subsampling.equilibrium_detection`.

    Parameters
    ----------
    df : DataFrame
        DataFrame to subsample according to the selected method.
    drop_duplicates : bool
        Drop the duplicated lines based on time.
    sort : bool
        Sort the Dataframe based on the time column.
    remove_burnin : bool
        Whether to perform equilibrium detection (``True``) or just do
        statistical inefficiency (``False``).

        .. versionadded:: 1.0.0

    **kwargs :
        Additional keyword arguments for
        :func:`~alchemlyb.preprocessing.subsampling.statistical_inefficiency`
        or :func:`~alchemlyb.preprocessing.subsampling.equilibrium_detection`.

    Returns
    -------
    DataFrame
        `df` subsampled.

    Note
    ----
    The default of ``True`` for `drop_duplicates` and `sort` should result in
    robust decorrelation but can loose data.


    .. versionadded:: 0.6.0
    .. versionchanged:: 1.0.0
       Add the `remove_burnin` keyword to allow unequilibrated frames to be
       removed.

    """

    kwargs["drop_duplicates"] = drop_duplicates
    kwargs["sort"] = sort

    series = dhdl2series(df)

    if remove_burnin:
        return equilibrium_detection(df, series, **kwargs)
    else:
        return statistical_inefficiency(df, series, **kwargs)


@pass_attrs
def u_nk2series(df, method="dE"):
    """Convert an u_nk DataFrame into a series based on the selected method
    for subsampling.

    The method can be either 'all' (obtained as a sum over all energy
    components) or 'dE'. In the latter case the energy differences
    :math:`dE_{i,i+1}` (:math:`dE_{i,i-1}` for the last lambda) are used.

    Parameters
    ----------
    df : DataFrame
        DataFrame to be converted according to the selected method.
    method : {'all', 'dE'}
        Method for converting the data.

    Returns
    -------
    Series
        `series` to be used as input for
        :func:`~alchemlyb.preprocessing.subsampling.statistical_inefficiency`
        or
        :func:`~alchemlyb.preprocessing.subsampling.equilibrium_detection`.


    .. versionadded:: 1.0.0

    """

    # deprecation: remove in 3.0.0
    # (the deprecations should show up in the calling functions)
    if method == "dhdl":
        warnings.warn(
            "Method 'dhdl' has been deprecated, using 'dE' instead. "
            "'dhdl' will be removed in alchemlyb 3.0.0.",
            category=DeprecationWarning,
            stacklevel=2,
        )
        method = "dE"
    elif method == "dhdl_all":
        warnings.warn(
            "Method 'dhdl_all' has been deprecated, using 'all' instead. "
            "'dhdl_all' will be removed in alchemlyb 3.0.0.",
            category=DeprecationWarning,
            stacklevel=2,
        )
        method = "all"

    # Check if the input is u_nk
    try:
        key = df.index.values[0][1:]
        if len(key) == 1:
            key = key[0]
        df[key]
    except KeyError:
        raise ValueError("The input should be u_nk")

    if method == "all":
        series = df.sum(axis=1)
    elif method == "dE":
        # Using the same logic as alchemical-analysis
        key = df.index.values[0][1:]
        if len(key) == 1:
            # For the case where there is a single lambda
            index = df.columns.values.tolist().index(key[0])
        else:
            # For the case of more than 1 lambda
            index = df.columns.values.tolist().index(key)
            # for the state that is not the last state, take the state+1
        if index + 1 < len(df.columns):
            series = df.iloc[:, index + 1]
            # for the state that is the last state, take the state-1
        else:
            series = df.iloc[:, index - 1]
    else:
        raise ValueError("Decorrelation method {} not found.".format(method))
    return series


@pass_attrs
def dhdl2series(df, method="all"):
    """Convert a dhdl DataFrame to a series for subsampling.

    The series is generated by summing over all energy components (axis 1 of
    `df`), as for ``method='all'`` in :func:`u_nk2series`. Commonly, `df` only
    contains a single energy component but in some cases (such as using a split
    protocol in GROMACS), it can contain multiple columns for different energy
    terms.

    Parameters
    ----------
    df : DataFrame
        DataFrame to subsample according to the selected method.
    method : 'all'
        Only 'all' is available; the keyword is provided for compatibility with
        :func:`u_nk2series`.

    Returns
    -------
    Series
        `series` to be used as input for
        :func:`~alchemlyb.preprocessing.subsampling.statistical_inefficiency`
        or
        :func:`~alchemlyb.preprocessing.subsampling.equilibrium_detection`.


    .. versionadded:: 1.0.0

    """
    if method != "all":
        raise ValueError("Only method='all' is supported for dhdl2series().")
    series = df.sum(axis=1)
    return series


def _check_multiple_times(df):
    if isinstance(df, pd.Series):
        return (
            df.sort_index(axis=0).reset_index("time", name="").duplicated("time").any()
        )
    else:
        return df.sort_index(axis=0).reset_index("time").duplicated("time").any()


def _check_sorted(df):
    return df.reset_index(0)["time"].is_monotonic_increasing


def _drop_duplicates(df, series=None):
    """Drop the duplication in the ``df`` which could be Dataframe or
    Series, if series is provided, format the series such that it has the
    same length as ``df``.

    Parameters
    ----------
    df : DataFrame or Series
        DataFrame or Series where duplication will be dropped.
    series : Series
        series to be formatted in the same way as df.

    Returns
    -------
    df : DataFrame or Series
        Formatted DataFrame or Series.
    series : Series
        Formatted Series.
    """
    if isinstance(df, pd.Series):
        # remove the duplicate based on time
        drop_duplicates_series = df.reset_index("time", name="").drop_duplicates("time")
        # Rest the time index
        lambda_names = [
            "time",
        ]
        lambda_names.extend(drop_duplicates_series.index.names)
        df = drop_duplicates_series.set_index("time", append=True).reorder_levels(
            lambda_names
        )
    else:
        # remove the duplicate based on time
        drop_duplicates_df = df.reset_index("time").drop_duplicates("time")
        # Rest the time index
        lambda_names = [
            "time",
        ]
        lambda_names.extend(drop_duplicates_df.index.names)
        df = drop_duplicates_df.set_index("time", append=True).reorder_levels(
            lambda_names
        )

    # Do the same withing with the series
    if series is not None:
        # remove the duplicate based on time
        drop_duplicates_series = series.reset_index("time", name="").drop_duplicates(
            "time"
        )
        # Rest the time index
        lambda_names = [
            "time",
        ]
        lambda_names.extend(drop_duplicates_series.index.names)
        series = drop_duplicates_series.set_index("time", append=True).reorder_levels(
            lambda_names
        )
    return df, series


def _sort_by_time(df, series=None):
    """Sort the ``df`` by time which could be Dataframe or
    Series, if series is provided, sort the series as well.

    Parameters
    ----------
    df : DataFrame or Series
        DataFrame or Series to be sorted by time.
    series : Series
        series to be sorted by time.

    Returns
    -------
    df : DataFrame or Series
        Formatted DataFrame or Series.
    series : Series
        Formatted Series.
    """
    df = df.sort_index(level="time")

    if series is not None:
        series = series.sort_index(level="time")
    return df, series


def _prepare_input(df, series, drop_duplicates, sort):
    """Prepare and check the input to be used for statistical_inefficiency or equilibrium_detection.

    Parameters
    ----------
    df : DataFrame or Series
        DataFrame or Series to be Prepared and checked.
    series : Series
        series to be Prepared and checked.

    Returns
    -------
    df : DataFrame or Series
        Formatted DataFrame or Series.
    series : Series
        Formatted Series.
    """
    if _check_multiple_times(df):
        if drop_duplicates:
            df, series = _drop_duplicates(df, series)
        else:
            raise KeyError(
                "Duplicate time values found; statistical inefficiency "
                "only works on a single, contiguous, "
                "and sorted timeseries."
            )

    if not _check_sorted(df):
        if sort:
            df, series = _sort_by_time(df, series)
        else:
            raise KeyError(
                "Statistical inefficiency only works as expected if "
                "values are sorted by time, increasing."
            )

    if series is not None:
        if len(series) != len(df) or not all(
            series.reset_index()["time"] == df.reset_index()["time"]
        ):
            raise ValueError("series and data must be sampled at the same times")
    return df, series


def slicing(df, lower=None, upper=None, step=None, force=False):
    """Subsample a DataFrame using simple slicing.

    Parameters
    ----------
    df : DataFrame
        DataFrame to subsample.
    lower : float
        Lower time to slice from.
    upper : float
        Upper time to slice to (inclusive).
    step : int
        Step between rows to slice by.
    force : bool
        Ignore checks that DataFrame is in proper form for expected behavior.

    Returns
    -------
    DataFrame
        `df` subsampled.


    .. versionchanged:: 1.0.1
       The rows with NaN values are not dropped by default.
    """
    try:
        df = df.loc[lower:upper:step]
    except KeyError:
        raise KeyError("DataFrame rows must be sorted by time, increasing.")

    if not force and _check_multiple_times(df):
        raise KeyError(
            "Duplicate time values found; it's generally advised "
            "to use slicing on DataFrames with unique time values "
            "for each row. Use `force=True` to ignore this error."
        )

    return df


def statistical_inefficiency(
    df,
    series=None,
    lower=None,
    upper=None,
    step=None,
    conservative=True,
    drop_duplicates=False,
    sort=False,
):
    """Subsample a DataFrame based on the calculated statistical inefficiency
    of a timeseries.

    If `series` is ``None``, then this function will behave the same as
    :func:`slicing`.

    Parameters
    ----------
    df : DataFrame
        DataFrame to subsample according statistical inefficiency of `series`.
    series : Series
        Series to use for calculating statistical inefficiency. If ``None``,
        no statistical inefficiency-based subsampling will be performed.
    lower : float
        Lower bound to pre-slice `series` data from.
    upper : float
        Upper bound to pre-slice `series` to (inclusive).
    step : int
        Step between `series` items to pre-slice by.
    conservative : bool
        ``True`` use ``ceil(statistical_inefficiency)`` to slice the data in uniform
        intervals (the default). ``False`` will sample at non-uniform intervals to
        closely match the (fractional) statistical_inefficieny, as implemented
        in :func:`pymbar.timeseries.subsample_correlated_data`.
    drop_duplicates : bool
        Drop the duplicated lines based on time.
    sort : bool
        Sort the Dataframe based on the time column.

    Returns
    -------
    DataFrame
        `df` subsampled according to subsampled `series`.

    Warning
    -------
    The `series` and the data to be sliced, `df`, need to have the same number
    of elements because the statistical inefficiency is calculated based on
    the index of the series (and not an associated time). At the moment there is
    no automatic conversion from a time to an index.

    Note
    ----
    For a non-integer statistical ineffciency :math:`g`, the default value
    ``conservative=True`` will provide _fewer_ data points than allowed by
    :math:`g` and thus error estimates will be _higher_. For large numbers of
    data points and converged free energies, the choice should not make a
    difference. For small numbers of data points, ``conservative=True``
    decreases a false sense of accuracy and is deemed the more careful and
    conservative approach.

    See Also
    --------
    pymbar.timeseries.statistical_inefficiency : detailed background
    pymbar.timeseries.subsample_correlated_data : used for subsampling


    .. versionchanged:: 0.2.0
       The ``conservative`` keyword was added and the method is now using
       ``pymbar.timeseries.statistical_inefficiency()``; previously, the statistical
       inefficiency was _rounded_ (instead of ``ceil()``) and thus one could
       end up with correlated data.

    .. versionchanged:: 1.0.0
       Fixed a bug that would effectively ignore the ``lower`` and ``step``
       keywords when returning the subsampled DataFrame object. See
       `issue #198 <https://github.com/alchemistry/alchemlyb/issues/198>`_ for
       more details.

    """
    df, series = _prepare_input(df, series, drop_duplicates, sort)

    if series is not None:
        series = slicing(series, lower=lower, upper=upper, step=step)
        df = slicing(df, lower=lower, upper=upper, step=step)

        # calculate statistical inefficiency of series (could use fft=True but needs test)
        statinef = _statistical_inefficiency(series)

        # use the subsample_correlated_data function to get the subsample index
        indices = _subsample_correlated_data(
            series, g=statinef, conservative=conservative
        )
        df = df.iloc[indices]
    else:
        df = slicing(df, lower=lower, upper=upper, step=step)

    return df


def equilibrium_detection(
    df,
    series=None,
    lower=None,
    upper=None,
    step=None,
    drop_duplicates=False,
    sort=False,
):
    """Subsample a DataFrame using automated equilibrium detection on a timeseries.

    This function uses the :mod:`pymbar` implementation of the *simple
    automated equilibrium detection* algorithm in [Chodera2016]_.

    If `series` is ``None``, then this function will behave the same as
    :func:`slicing`.

    Parameters
    ----------
    df : DataFrame
        DataFrame to subsample according to equilibrium detection on `series`.
    series : Series
        Series to detect equilibration on. If ``None``, no equilibrium
        detection-based subsampling will be performed.
    lower : float
        Lower bound to pre-slice `series` data from.
    upper : float
        Upper bound to pre-slice `series` to (inclusive).
    step : int
        Step between `series` items to pre-slice by.
    drop_duplicates : bool
        Drop the duplicated lines based on time.
    sort : bool
        Sort the Dataframe based on the time column.

    Returns
    -------
    DataFrame
        `df` subsampled according to subsampled `series`.

    Notes
    -----
    Please cite [Chodera2016]_ when you use this function in published work.

    See Also
    --------
    pymbar.timeseries.detect_equilibration : detailed background
    pymbar.timeseries.subsample_correlated_data : used for subsampling


    .. versionchanged:: 1.0.0
        Add the drop_duplicates and sort keyword to unify the behaviour between
        :func:`~alchemlyb.preprocessing.subsampling.statistical_inefficiency` or
        :func:`~alchemlyb.preprocessing.subsampling.equilibrium_detection`.

    """
    df, series = _prepare_input(df, series, drop_duplicates, sort)

    if series is not None:
        series = slicing(series, lower=lower, upper=upper, step=step)
        df = slicing(df, lower=lower, upper=upper, step=step)

        # calculate statistical inefficiency of series, with equilibrium detection
        t, statinef, Neff_max = _detect_equilibration(series.values)

        series_equil = series[t:]
        df_equil = df[t:]

        indices = _subsample_correlated_data(series_equil, g=statinef)
        df = df_equil.iloc[indices]
    else:
        df = slicing(df, lower=lower, upper=upper, step=step)

    return df



class _EstimatorMixOut:
    """This class creates view for the d_delta_f_, delta_f_, states_ for the
    estimator class to consume."""

    _d_delta_f_ = None
    _delta_f_ = None
    _states_ = None

    @property
    def d_delta_f_(self):
        return self._d_delta_f_

    @property
    def delta_f_(self):
        return self._delta_f_

    @property
    def states_(self):
        return self._states_



import logging

import pandas as pd
import pymbar
from sklearn.base import BaseEstimator

from .base import _EstimatorMixOut


class MBAR(BaseEstimator, _EstimatorMixOut):
    r"""Multi-state Bennett acceptance ratio (MBAR).

    Parameters
    ----------

    maximum_iterations : int, optional
        Set to limit the maximum number of iterations performed.

    relative_tolerance : float, optional
        Set to determine the relative tolerance convergence criteria.

    initial_f_k : np.ndarray, float, shape=(K), optional
        Set to the initial dimensionless free energies to use as a
        guess (default None, which sets all :math:`f_k = 0`).

    method : str, optional, default="robust"
        The optimization routine to use.  This can be any of the methods
        available via :func:`scipy.optimize.minimize` or 
        :func:`scipy.optimize.root`.

    verbose : bool, optional
        Set to ``True`` if verbose debug output from :mod:`pymbar` is desired.
        Output from alchemlyb is logged via :mod:`logging`.

    Attributes
    ----------

    delta_f_ : DataFrame
        The estimated dimensionless free energy difference between each state.

    d_delta_f_ : DataFrame
        The estimated statistical uncertainty (one standard deviation) in
        dimensionless free energy differences.

    theta_ : DataFrame
        The theta matrix.

    states_ : list
        Lambda states for which free energy differences were obtained.

    Notes
    -----
    See [Shirts2008]_ for details of the derivation and cite the
    paper when using MBAR in published work.

    See Also
    --------
    pymbar.mbar.MBAR


    .. versionchanged:: 1.0.0
       `delta_f_`, `d_delta_f_`, `states_` are view of the original object.
    .. versionchanged:: 2.0.0
        default value for `method` was changed from "hybr" to "robust"
    """

    def __init__(
        self,
        maximum_iterations=10000,
        relative_tolerance=1.0e-7,
        initial_f_k=None,
        method="robust",
        verbose=False,
    ):
        self.maximum_iterations = maximum_iterations
        self.relative_tolerance = relative_tolerance
        self.initial_f_k = initial_f_k
        self.method = method
        self.verbose = verbose
        self.logger = logging.getLogger("alchemlyb.estimators.MBAR")

        # handle for pymbar.MBAR object
        self._mbar = None

    def fit(self, u_nk):
        """
        Compute overlap matrix of reduced potentials using multi-state
        Bennett acceptance ratio.

        Parameters
        ----------
        u_nk : DataFrame
            ``u_nk[n, k]`` is the reduced potential energy of uncorrelated
            configuration ``n`` evaluated at state ``k``.

        """
        # sort by state so that rows from same state are in contiguous blocks
        u_nk = u_nk.sort_index(level=u_nk.index.names[1:])

        groups = u_nk.groupby(level=u_nk.index.names[1:])
        N_k = [
            (len(groups.get_group(i)) if i in groups.groups else 0)
            for i in u_nk.columns
        ]
        self._states_ = u_nk.columns.values.tolist()

        self._mbar = pymbar.MBAR(
            u_nk.T,
            N_k,
            maximum_iterations=self.maximum_iterations,
            relative_tolerance=self.relative_tolerance,
            verbose=self.verbose,
            initial_f_k=self.initial_f_k,
            solver_protocol=self.method,
        )
        out = self._mbar.compute_free_energy_differences(return_theta=True)
        self._delta_f_ = pd.DataFrame(
            out["Delta_f"], columns=self._states_, index=self._states_
        )
        self._d_delta_f_ = pd.DataFrame(
            out["dDelta_f"], columns=self._states_, index=self._states_
        )
        self.theta_ = pd.DataFrame(
            out["Theta"], columns=self._states_, index=self._states_
        )

        self._delta_f_.attrs = u_nk.attrs
        self._d_delta_f_.attrs = u_nk.attrs

        return self

    @property
    def overlap_matrix(self):
        r"""MBAR overlap matrix.

        The estimated state overlap matrix :math:`O_{ij}` is an estimate of the probability
        of observing a sample from state :math:`i` in state :math:`j`.

        The :attr:`overlap_matrix` is computed on-the-fly. Assign it to a variable if
        you plan to re-use it.

        See Also
        ---------
        pymbar.mbar.MBAR.computeOverlap
        """
        return self._mbar.compute_overlap()["matrix"]



"""Functions for assessing convergence of free energy estimates and raw data."""

import logging
from warnings import warn

import numpy as np
import pandas as pd

from .. import concat
from ..estimators import BAR, TI, MBAR, FEP_ESTIMATORS, TI_ESTIMATORS
from ..postprocessors.units import to_kT

estimators_dispatch = {"BAR": BAR, "TI": TI, "MBAR": MBAR}


def forward_backward_convergence(df_list, estimator="MBAR", num=10, **kwargs):
    """Forward and backward convergence of the free energy estimate.

    Generate the free energy estimate as a function of time in both directions,
    with the specified number of equally spaced points in the time
    [Klimovich2015]_. For example, setting `num` to 10 would give the forward
    convergence which is the free energy estimate from the first 10%, 20%, 30%,
    ... of the data. The Backward would give the estimate from the last 10%,
    20%, 30%, ... of the data.

    Parameters
    ----------
    df_list : list
        List of DataFrame of either dHdl or u_nk.
    estimator : {'MBAR', 'BAR', 'TI'}
        Name of the estimators.
        See the important note below on the use of "MBAR".

        .. deprecated:: 1.0.0
           Lower case input is also accepted until release 2.0.0.
    num : int
        The number of time points.
    kwargs : dict
        Keyword arguments to be passed to the estimator.

    Returns
    -------
    :class:`pandas.DataFrame`
        The DataFrame with convergence data. ::

                Forward  Forward_Error  Backward  Backward_Error  data_fraction
            0  3.016442       0.052748  3.065176        0.051036            0.1
            1  3.078106       0.037170  3.078567        0.036640            0.2
            2  3.072561       0.030186  3.047357        0.029775            0.3
            3  3.048325       0.026070  3.057527        0.025743            0.4
            4  3.049769       0.023359  3.037454        0.023001            0.5
            5  3.034078       0.021260  3.040484        0.021075            0.6
            6  3.043274       0.019642  3.032495        0.019517            0.7
            7  3.035460       0.018340  3.036670        0.018261            0.8
            8  3.042032       0.017319  3.046597        0.017233            0.9
            9  3.044149       0.016405  3.044385        0.016402            1.0


    .. versionadded:: 0.6.0
    .. versionchanged:: 1.0.0
       The ``estimator`` accepts uppercase input.
       The default for using ``estimator='MBAR'`` was changed from
       :class:`~alchemlyb.estimators.MBAR` to :class:`~alchemlyb.estimators.AutoMBAR`.
    .. versionchanged:: 2.0.0
       Use pymbar.MBAR instead of the AutoMBAR option.

    """
    logger = logging.getLogger("alchemlyb.convergence." "forward_backward_convergence")
    logger.info("Start convergence analysis.")
    logger.info("Check data availability.")
    if estimator.upper() != estimator:
        warn(
            "Using lower-case strings for the 'estimator' kwarg in "
            "convergence.forward_backward_convergence() is deprecated in "
            "1.0.0 and only upper case will be accepted in 2.0.0",
            DeprecationWarning,
        )
        estimator = estimator.upper()

    if estimator not in (FEP_ESTIMATORS + TI_ESTIMATORS):
        msg = f"Estimator {estimator} is not available in {FEP_ESTIMATORS + TI_ESTIMATORS}."
        logger.error(msg)
        raise ValueError(msg)
    else:
        # select estimator class by name
        estimator_fit = estimators_dispatch[estimator](**kwargs).fit
        logger.info(f"Use {estimator} estimator for convergence analysis.")

    logger.info("Begin forward analysis")
    forward_list = []
    forward_error_list = []
    for i in range(1, num + 1):
        logger.info("Forward analysis: {:.2f}%".format(100 * i / num))
        sample = []
        for data in df_list:
            sample.append(data[: len(data) // num * i])
        sample = concat(sample)
        result = estimator_fit(sample)
        forward_list.append(result.delta_f_.iloc[0, -1])
        if estimator.lower() == "bar":
            error = np.sqrt(
                sum(
                    [
                        result.d_delta_f_.iloc[i, i + 1] ** 2
                        for i in range(len(result.d_delta_f_) - 1)
                    ]
                )
            )
            forward_error_list.append(error)
        else:
            forward_error_list.append(result.d_delta_f_.iloc[0, -1])
        logger.info(
            "{:.2f} +/- {:.2f} kT".format(forward_list[-1], forward_error_list[-1])
        )

    logger.info("Begin backward analysis")
    backward_list = []
    backward_error_list = []
    for i in range(1, num + 1):
        logger.info("Backward analysis: {:.2f}%".format(100 * i / num))
        sample = []
        for data in df_list:
            sample.append(data[-len(data) // num * i :])
        sample = concat(sample)
        result = estimator_fit(sample)
        backward_list.append(result.delta_f_.iloc[0, -1])
        if estimator.lower() == "bar":
            error = np.sqrt(
                sum(
                    [
                        result.d_delta_f_.iloc[i, i + 1] ** 2
                        for i in range(len(result.d_delta_f_) - 1)
                    ]
                )
            )
            backward_error_list.append(error)
        else:
            backward_error_list.append(result.d_delta_f_.iloc[0, -1])
        logger.info(
            "{:.2f} +/- {:.2f} kT".format(backward_list[-1], backward_error_list[-1])
        )

    convergence = pd.DataFrame(
        {
            "Forward": forward_list,
            "Forward_Error": forward_error_list,
            "Backward": backward_list,
            "Backward_Error": backward_error_list,
            "data_fraction": [i / num for i in range(1, num + 1)],
        }
    )
    convergence.attrs = df_list[0].attrs
    return convergence


def _cummean(vals, out_length):
    """The cumulative mean of an array.

    This function computes the cumulative mean and shapes the result to the
    desired length.

    Parameters
    ----------
    vals : numpy.array
        The one-dimensional input array.
    out_length : int
        The length of the output array.

    Returns
    -------
    numpy.array
        The one-dimensional input array with length of total.

    Note
    ----
    If the length of the input series is smaller than the ``out_length``, the
    length of the output array is the same as the input series.


    .. versionadded:: 1.0.0

    """
    in_length = len(vals)
    if in_length < out_length:
        out_length = in_length
    block = in_length // out_length
    reshape = vals[: block * out_length].reshape(block, out_length)
    mean = np.mean(reshape, axis=0)
    result = np.cumsum(mean) / np.arange(1, out_length + 1)
    return result


def fwdrev_cumavg_Rc(series, precision=0.01, tol=2):
    r"""Generate the convergence criteria :math:`R_c` for a single simulation.

    The input will be :class:`pandas.Series` generated by
    :func:`~alchemlyb.preprocessing.subsampling.decorrelate_u_nk` or
    :func:`~alchemlyb.preprocessing.subsampling.decorrelate_dhdl`.

    The output will be the float :math:`R_c` [Fan2020]_ [Fan2021]_ and a
    :class:`pandas.DataFrame` with the forward and backward cumulative average
    at `precision` fractional increments, as described below.

    :math:`R_c = 0` indicates that the system is well equilibrated right from
    the beginning while :math:`R_c = 1` signifies that the whole trajectory is
    not equilibrated.

    Parameters
    ----------
    series : pandas.Series
        The input energy array.
    precision : float
        The precision of the output :math:`R_c`. To speed the calculation up,
        the data has been block-averaged before doing the calculation, the size
        of the block is controlled by the desired precision.
    tol : float
        Tolerance (or convergence threshold :math:`\epsilon` in [Fan2021]_)
        in :math:`kT`.

    Returns
    -------
    float
        Convergence time fraction :math:`R_c` [Fan2021]_
    :class:`pandas.DataFrame`
        The DataFrame with moving average. ::

                Forward  Backward  data_fraction
            0  3.016442  3.065176            0.1
            1  3.078106  3.078567            0.2
            2  3.072561  3.047357            0.3
            3  3.048325  3.057527            0.4
            4  3.049769  3.037454            0.5
            5  3.034078  3.040484            0.6
            6  3.043274  3.032495            0.7
            7  3.035460  3.036670            0.8
            8  3.042032  3.046597            0.9
            9  3.044149  3.044385            1.0

    Notes
    -----
    This function computes :math:`R_c` from `equation 16`_ from [Fan2021]_.
    The code is modified based on Shujie Fan's (@VOD555) work. Zhiyi Wu
    (@xiki-tempula) improved the performance of the original algorithm.

    Please cite [Fan2021]_ when using this function.

    See also
    --------
    A_c


    .. versionadded:: 1.0.0

    .. _`equation 16`:
       https://www.ncbi.nlm.nih.gov/pmc/articles/PMC8397498/#FD16

    """
    series = to_kT(series)
    array = series.to_numpy()
    out_length = int(1 / precision)
    g_forward = _cummean(array, out_length)
    g_backward = _cummean(array[::-1], out_length)
    length = len(g_forward)

    convergence = pd.DataFrame(
        {
            "Forward": g_forward,
            "Backward": g_backward,
            "data_fraction": [i / length for i in range(1, length + 1)],
        }
    )
    convergence.attrs = series.attrs

    # Final value
    g = g_forward[-1]
    conv = np.logical_and(np.abs(g_forward - g) < tol, np.abs(g_backward - g) < tol)
    for i in range(out_length):
        if all(conv[i:]):
            return i / length, convergence
    else:
        # This branch exists as we are truncating the dataset to speed the
        # calculation up. For example if the dataset has 21 points and the
        # precision is 0.05. The last point of g_forward is computed using
        # data[0:20] while the last point of g_backward is computed using
        # data[1:21]. Thus, the last point of g_forward and g_backward are not
        # the same as this branch will be triggered.
        return 1.0, convergence


def A_c(series_list, precision=0.01, tol=2):
    r"""Generate the ensemble convergence criteria :math:`A_c` for a set of simulations.

    The input is a :class:`list` of :class:`pandas.Series` generated by
    :func:`~alchemlyb.preprocessing.subsampling.decorrelate_u_nk` or
    :func:`~alchemlyb.preprocessing.subsampling.decorrelate_dhdl`.

    The output will the float :math:`A_c` [Fan2020]_ [Fan2021]_. :math:`A_c` is
    a number between 0 and 1 that can be interpreted as the ratio of the total
    equilibrated simulation time to the whole simulation time for a full set of
    simulations. :math:`A_c = 1` means that all simulation time frames in all
    windows can be considered equilibrated, while :math:`A_c = 0` indicates
    that nothing is equilibrated.

    Parameters
    ----------
    series_list : list
        A list of :class:`pandas.Series` energy array.
    precision : float
        The precision of the output :math:`A_c`. To speed the calculation up, the data
        has been block-averaged before doing the calculation, the size of the
        block is controlled by the desired precision.
    tol : float
        Tolerance (or convergence threshold :math:`\epsilon` in [Fan2021]_)
        in :math:`kT`.

    Returns
    -------
    float
        The area :math:`A_c` under curve for convergence time fraction.

    Notes
    -----
    This function computes :math:`A_c` from `equation 18`_ from [Fan2021]_.

    Please cite [Fan2021]_ when using this function.

    See also
    --------
    fwdrev_cumavg_Rc


    .. versionadded:: 1.0.0

    .. _`equation 18`:
       https://www.ncbi.nlm.nih.gov/pmc/articles/PMC8397498/#FD18

    """
    logger = logging.getLogger("alchemlyb.convergence.A_c")
    n_R_c = len(series_list)
    R_c_list = [fwdrev_cumavg_Rc(series, precision, tol)[0] for series in series_list]
    logger.info(f"R_c list: {R_c_list}")
    # Integrate the R_c_list <= R_c over the range of 0 to 1
    array_01 = np.hstack((R_c_list, [0, 1]))
    sorted_array = np.sort(np.unique(array_01))
    result = 0
    for i, element in enumerate(sorted_array[::-1]):
        if i == 0:
            continue
        else:
            d_R_c = sorted_array[-i] - sorted_array[-i - 1]
            result += d_R_c * sum(R_c_list <= element) / n_R_c
    return result



"""Parsers for extracting alchemical data from `GOMC <http://gomc.eng.wayne.edu/>`_ output files.

"""
import pandas as pd

from . import _init_attrs
from .util import anyopen
from ..postprocessors.units import R_kJmol

k_b = R_kJmol


@_init_attrs
def extract_u_nk(filename, T):
    """Return reduced potentials `u_nk` from a Hamiltonian differences dat file.

    Parameters
    ----------
    filename : str
        Path to free energy file to extract data from.
    T : float
        Temperature in Kelvin at which the simulation was sampled.

    Returns
    -------
    u_nk : DataFrame
        Potential energy for each alchemical state (k) for each frame (n).


    .. versionchanged:: 0.5.0
        The :mod:`scipy.constants` is used for parsers instead of
        the constants used by the corresponding MD engine.

    """

    dh_col_match = "dU/dL"
    h_col_match = "DelE"
    pv_col_match = "PV"
    u_col_match = ["Total_En"]
    beta = 1 / (k_b * T)

    state, lambdas, statevec = _extract_state(filename)

    # extract a DataFrame from free energy file data
    df = _extract_dataframe(filename)

    times = df[df.columns[0]]

    # want to grab only dH columns
    DHcols = [col for col in df.columns if (h_col_match in col)]
    dH = df[DHcols]

    # GOMC also gives us pV directly; need this for reduced potential
    pv_cols = [col for col in df.columns if (pv_col_match in col)]
    pv = None
    if pv_cols:
        pv = df[pv_cols[0]]

    # GOMC also gives us total energy U directly; need this for reduced potential
    u_cols = [
        col
        for col in df.columns
        if any(single_u_col_match in col for single_u_col_match in u_col_match)
    ]
    u = None
    if u_cols:
        u = df[u_cols[0]]

    u_k = dict()
    cols = list()
    for col in dH:
        u_col = eval(col.split("->")[1][:-1])
        # calculate reduced potential u_k = dH + pV + U
        u_k[u_col] = beta * dH[col].values
        if pv_cols:
            u_k[u_col] += beta * pv.values
        if u_cols:
            u_k[u_col] += beta * u.values
        cols.append(u_col)

    u_k = pd.DataFrame(
        u_k, columns=cols, index=pd.Index(times.values, name="time", dtype="Float64")
    )

    # Need to modify the lambda name
    cols = [l + "-lambda" for l in lambdas]
    # create columns for each lambda, indicating state each row sampled from
    for i, l in enumerate(cols):
        u_k[l] = statevec[i]

    # set up new multi-index
    newind = ["time"] + cols
    u_k = u_k.reset_index().set_index(newind)

    u_k.name = "u_nk"

    return u_k


@_init_attrs
def extract_dHdl(filename, T):
    """Return gradients `dH/dl` from a Hamiltonian differences free energy file.

    Parameters
    ----------
    filename : str
        Path to free energy file to extract data from.
    T : float
        Temperature in Kelvin at which the simulation was sampled.

    Returns
    -------
    dH/dl : Series
        dH/dl as a function of step for this lambda window.


    .. versionchanged:: 0.5.0
        The :mod:`scipy.constants` is used for parsers instead of
        the constants used by the corresponding MD engine.

    """
    beta = 1 / (k_b * T)

    state, lambdas, statevec = _extract_state(filename)

    # extract a DataFrame from free energy data
    df = _extract_dataframe(filename)

    times = df[df.columns[0]]

    # want to grab only dH/dl columns
    dHcols = []
    for l in lambdas:
        dHcols.extend([col for col in df.columns if (l in col)])

    dHdl = df[dHcols]

    # make dimensionless
    dHdl *= beta

    dHdl = pd.DataFrame(
        dHdl.values,
        columns=lambdas,
        index=pd.Index(times.values, name="time", dtype="Float64"),
    )

    # Need to modify the lambda name
    cols = [l + "-lambda" for l in lambdas]
    # create columns for each lambda, indicating state each row sampled from
    for i, l in enumerate(cols):
        dHdl[l] = statevec[i]

    # set up new multi-index
    newind = ["time"] + cols
    dHdl = dHdl.reset_index().set_index(newind)

    dHdl.name = "dH/dl"

    return dHdl


def extract(filename, T):
    r"""Return reduced potentials `u_nk` and gradients `dH/dl`
    from a Hamiltonian differences free energy file.

    Parameters
    ----------
    xvg : str
        Path to free energy file to extract data from.
    T : float
        Temperature in Kelvin the simulations sampled.
    filter : bool
        Filter out the lines that cannot be parsed.
        Such as rows with incorrect number of Columns and incorrectly
        formatted numbers (e.g. 123.45.67, nan or -).

    Returns
    -------
    Dict
        A dictionary with keys of 'u_nk', which is a pandas DataFrame of
        potential energy for each alchemical state (k) for each frame (n),
        and 'dHdl', which is a Series of dH/dl
        as a function of time for this lambda window.


    .. versionadded:: 1.0.0
    """

    return {"u_nk": extract_u_nk(filename, T), "dHdl": extract_dHdl(filename, T)}


def _extract_state(filename):
    """Extract information on state sampled, names of lambdas."""
    state = None
    with anyopen(filename, "r") as f:
        for line in f:
            if ("#" in line) and ("State" in line):
                state = int(line.split("State")[1].split(":")[0])
                # GOMC always print these two fields
                lambdas = ["Coulomb", "VDW"]
                statevec = eval(line.strip().split(" = ")[-1])
                break

    return state, lambdas, statevec


def _extract_dataframe(filename):
    """Extract a DataFrame from free energy data."""
    dh_col_match = "dU/dL"
    h_col_match = "DelE"
    pv_col_match = "PV"
    u_col_match = "Total_En"

    xaxis = "time"
    with anyopen(filename, "r") as f:
        names = []
        rows = []
        for line in f:
            line = line.strip()
            if len(line) == 0:
                # avoid parsing empty line
                continue
            elif line.startswith("#T"):
                # this line has state information. No need to be parsed
                continue
            elif line.startswith("#Steps"):
                # read the headers
                elements = line.split()
                for i, element in enumerate(elements):
                    if element.startswith(u_col_match):
                        names.append(element)
                    elif element.startswith(dh_col_match):
                        names.append(element)
                    elif element.startswith(h_col_match):
                        names.append(element)
                    elif element.startswith(pv_col_match):
                        names.append(element)
            else:
                # parse line as floats
                row = map(float, line.split())
                rows.append(row)

    cols = [xaxis]
    cols.extend(names)

    return pd.DataFrame(rows, columns=cols)



import numpy as np
import pandas as pd
from pymbar.other_estimators import bar as BAR_
from sklearn.base import BaseEstimator

from .base import _EstimatorMixOut


class BAR(BaseEstimator, _EstimatorMixOut):
    """Bennett acceptance ratio (BAR).

    Parameters
    ----------

    maximum_iterations : int, optional
        Set to limit the maximum number of iterations performed.

    relative_tolerance : float, optional
        Set to determine the relative tolerance convergence criteria.

    method : str, optional, default='false-position'
        choice of method to solve BAR nonlinear equations,
        one of 'self-consistent-iteration' or 'false-position' (default: 'false-position')

    verbose : bool, optional
        Set to True if verbose debug output is desired.

    Attributes
    ----------

    delta_f_ : DataFrame
        The estimated dimensionless free energy difference between each state.

    d_delta_f_ : DataFrame
        The estimated statistical uncertainty (one standard deviation) in
        dimensionless free energy differences.

    states_ : list
        Lambda states for which free energy differences were obtained.

    Notes
    -----
    See [Bennett1976]_ for details of the derivation and cite the paper
    (together with [Shirts2008]_ for the Python implementation in
    :mod:`pymbar`) when using BAR in published work.

    When possible, use MBAR instead of BAR as it makes better use of the
    available data.

    See Also
    --------
    MBAR


    .. versionchanged:: 1.0.0
       `delta_f_`, `d_delta_f_`, `states_` are view of the original object.

    """

    def __init__(
        self,
        maximum_iterations=10000,
        relative_tolerance=1.0e-7,
        method="false-position",
        verbose=False,
    ):

        self.maximum_iterations = maximum_iterations
        self.relative_tolerance = relative_tolerance
        self.method = method
        self.verbose = verbose

        # handle for pymbar.BAR object
        self._bar = None

    def fit(self, u_nk):
        """
        Compute overlap matrix of reduced potentials using
        Bennett acceptance ratio.

        Parameters
        ----------
        u_nk : DataFrame
            u_nk[n,k] is the reduced potential energy of uncorrelated
            configuration n evaluated at state k.

        """
        # sort by state so that rows from same state are in contiguous blocks
        u_nk = u_nk.sort_index(level=u_nk.index.names[1:])

        # get a list of the lambda states
        self._states_ = u_nk.columns.values.tolist()

        # group u_nk by lambda states
        groups = u_nk.groupby(level=u_nk.index.names[1:])
        N_k = [
            (len(groups.get_group(i)) if i in groups.groups else 0)
            for i in u_nk.columns
        ]

        # Now get free energy differences and their uncertainties for each step
        deltas = np.array([])
        d_deltas = np.array([])
        for k in range(len(N_k) - 1):
            # get us from lambda step k
            uk = groups.get_group(self._states_[k])
            # get w_F
            w_f = uk.iloc[:, k + 1] - uk.iloc[:, k]

            # get us from lambda step k+1
            uk1 = groups.get_group(self._states_[k + 1])
            # get w_R
            w_r = uk1.iloc[:, k] - uk1.iloc[:, k + 1]

            # now determine df and ddf using pymbar.BAR
            out = BAR_(
                w_f,
                w_r,
                method=self.method,
                maximum_iterations=self.maximum_iterations,
                relative_tolerance=self.relative_tolerance,
                verbose=self.verbose,
            )

            df, ddf = out["Delta_f"], out["dDelta_f"]
            deltas = np.append(deltas, df)
            d_deltas = np.append(d_deltas, ddf**2)

        # build matrix of deltas between each state
        adelta = np.zeros((len(deltas) + 1, len(deltas) + 1))
        ad_delta = np.zeros_like(adelta)

        for j in range(len(deltas)):
            out = []
            dout = []
            for i in range(len(deltas) - j):
                out.append(deltas[i : i + j + 1].sum())

                # See https://github.com/alchemistry/alchemlyb/pull/60#issuecomment-430720742
                # Error estimate generated by BAR ARE correlated

                # Use the BAR uncertainties between two neighbour states
                if j == 0:
                    dout.append(d_deltas[i : i + j + 1].sum())
                # Other uncertainties are unknown at this point
                else:
                    dout.append(np.nan)

            adelta += np.diagflat(np.array(out), k=j + 1)
            ad_delta += np.diagflat(np.array(dout), k=j + 1)

        # yield standard delta_f_ free energies between each state
        self._delta_f_ = pd.DataFrame(
            adelta - adelta.T, columns=self._states_, index=self._states_
        )

        # yield standard deviation d_delta_f_ between each state
        self._d_delta_f_ = pd.DataFrame(
            np.sqrt(ad_delta + ad_delta.T), columns=self._states_, index=self._states_
        )
        self._delta_f_.attrs = u_nk.attrs
        self._d_delta_f_.attrs = u_nk.attrs

        return self



"""Tests for all FEP-based estimators in ``alchemlyb``.

"""
import pytest

import alchemlyb
from alchemlyb.estimators import MBAR, BAR


class FEPestimatorMixin:
    """Mixin for all FEP Estimator test classes."""

    def compare_delta_f(self, X_delta_f):
        est = self.cls().fit(X_delta_f[0])
        delta_f, d_delta_f = self.get_delta_f(est)

        assert X_delta_f[1] == pytest.approx(delta_f, rel=1e-3)
        assert X_delta_f[2] == pytest.approx(d_delta_f, rel=1e-3)

    def get_delta_f(self, est):
        # Use .iloc[0, -1] as we want to cater for both
        # delta_f_.loc[0.0, 1.0] and delta_f_.loc[(0.0, 0.0), (0.0, 1.0)]
        return est.delta_f_.iloc[0, -1], est.d_delta_f_.iloc[0, -1]


class TestMBAR(FEPestimatorMixin):
    """Tests for MBAR."""

    cls = MBAR

    @pytest.fixture(
        params=[
            ("gmx_benzene_Coulomb_u_nk", 3.041, 0.02088),
            ("gmx_benzene_VDW_u_nk", -3.007, 0.04519),
            ("gmx_expanded_ensemble_case_1", 75.923, 0.14124),
            ("gmx_expanded_ensemble_case_2", 75.915, 0.14372),
            ("gmx_expanded_ensemble_case_3", 76.173, 0.11345),
            ("gmx_water_particle_with_total_energy", -11.680, 0.083655),
            ("gmx_water_particle_with_potential_energy", -11.675, 0.083589),
            ("gmx_water_particle_without_energy", -11.654, 0.083415),
            ("amber_bace_example_complex_vdw", 2.41149, 0.0620658),
            ("gomc_benzene_u_nk", -0.79994, 0.091579),
        ],
    )
    def X_delta_f(self, request):
        get_unk, E, dE = request.param
        return alchemlyb.concat(request.getfixturevalue(get_unk)), E, dE

    def test_mbar(self, X_delta_f):
        self.compare_delta_f(X_delta_f)


class TestBAR(FEPestimatorMixin):
    """Tests for BAR."""

    cls = BAR

    @pytest.fixture(
        params=[
            ("gmx_benzene_Coulomb_u_nk", 3.044, 0.01640),
            ("gmx_benzene_VDW_u_nk", -3.033, 0.03438),
            ("gmx_expanded_ensemble_case_1", 75.993, 0.11056),
            ("gmx_expanded_ensemble_case_2", 76.009, 0.11220),
            ("gmx_expanded_ensemble_case_3", 76.219, 0.08886),
            ("gmx_water_particle_with_total_energy", -11.675, 0.065055),
            ("gmx_water_particle_with_potential_energy", -11.724, 0.064964),
            ("gmx_water_particle_without_energy", -11.660, 0.064914),
            ("amber_bace_example_complex_vdw", 2.39294, 0.051192),
            ("namd_tyr2ala", 11.0044, 0.10235),
            ("namd_idws", 0.221147, 0.041003),
            ("namd_idws_restarted", 7.081127, 0.0344211),
            ("namd_idws_restarted_reversed", -4.18405, 0.03457),
            ("gomc_benzene_u_nk", -0.87095, 0.071263),
        ],
    )
    def X_delta_f(self, request):
        get_unk, E, dE = request.param
        return alchemlyb.concat(request.getfixturevalue(get_unk)), E, dE

    def test_bar(self, X_delta_f):
        self.compare_delta_f(X_delta_f)

    def get_delta_f(self, est):
        ee = 0.0

        for i in range(len(est.d_delta_f_) - 1):
            ee += est.d_delta_f_.values[i][i + 1] ** 2
        # Use .iloc[0, -1] as we want to cater for both
        # delta_f_.loc[0.0, 1.0] and delta_f_.loc[(0.0, 0.0), (0.0, 1.0)]
        return est.delta_f_.iloc[0, -1], ee**0.5


class Test_Units:
    """Test the units."""

    @staticmethod
    @pytest.fixture()
    def u_nk(gmx_benzene_Coulomb_u_nk):
        return alchemlyb.concat(gmx_benzene_Coulomb_u_nk)

    def test_bar(self, u_nk):
        bar = BAR().fit(u_nk)
        assert bar.delta_f_.attrs["temperature"] == 300
        assert bar.delta_f_.attrs["energy_unit"] == "kT"
        assert bar.d_delta_f_.attrs["temperature"] == 300
        assert bar.d_delta_f_.attrs["energy_unit"] == "kT"

    def test_mbar(self, u_nk):
        mbar = MBAR().fit(u_nk)
        assert mbar.delta_f_.attrs["temperature"] == 300
        assert mbar.delta_f_.attrs["energy_unit"] == "kT"
        assert mbar.d_delta_f_.attrs["temperature"] == 300
        assert mbar.d_delta_f_.attrs["energy_unit"] == "kT"


class TestEstimatorMixOut:
    """Ensure that the attribute d_delta_f_, delta_f_, states_ cannot be
    modified."""

    @pytest.mark.parametrize("estimator", [MBAR, BAR])
    def test_d_delta_f_(self, estimator):
        _estimator = estimator()
        with pytest.raises(AttributeError):
            _estimator.d_delta_f_ = 1

    @pytest.mark.parametrize("estimator", [MBAR, BAR])
    def test_delta_f_(self, estimator):
        _estimator = estimator()
        with pytest.raises(AttributeError):
            _estimator.delta_f_ = 1

    @pytest.mark.parametrize("estimator", [MBAR, BAR])
    def test_states_(self, estimator):
        _estimator = estimator()
        with pytest.raises(AttributeError):
            _estimator.states_ = 1



"""Functions for Plotting the dhdl for the TI estimator.

To assess the quality of the TI estimator, the dhdl from lambda state 0
to lambda state 1 can plotted to assess if the change in dhdl is sampled
thoroughly.

The code for producing the dhdl plot is modified based on
`Alchemical Analysis <https://github.com/MobleyLab/alchemical-analysis>`_.

"""

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.font_manager import FontProperties as FP

from ..postprocessors.units import get_unit_converter


def plot_ti_dhdl(dhdl_data, labels=None, colors=None, units=None, ax=None):
    """Plot the dhdl of TI.

    Parameters
    ----------
    dhdl_data : :class:`~alchemlyb.estimators.TI` or list
        One or more :class:`~alchemlyb.estimators.TI` estimator, where the
        dhdl value will be taken from.
    labels : List
        list of labels for labelling all the alchemical transformations.
    colors : List
        list of colors for plotting all the alchemical transformations.
        Default: ['r', 'g', '#7F38EC', '#9F000F', 'b', 'y']
    units : str
        The unit of the estimate. The default is `None`, which is to use the
        unit in the input. Setting this will change the output unit.
    ax : matplotlib.axes.Axes
        Matplotlib axes object where the plot will be drawn on. If ``ax=None``,
        a new axes will be generated.

    Returns
    -------
    matplotlib.axes.Axes
        An axes with the TI dhdl drawn.

    Note
    ----
    The code is taken and modified from
    `Alchemical Analysis <https://github.com/MobleyLab/alchemical-analysis>`_.


    .. versionchanged:: 1.0.0
        If no units is given, the `units` in the input will be used.

    .. versionchanged:: 0.5.0
        The `units` will be used to change the underlying data instead of only
        changing the figure legend.

    .. versionadded:: 0.4.0
    """
    # Make it into a list
    # separate_dhdl method is used so that the input for the actual plotting
    # Function are a uniformed list of series object which only contains one
    # lambda.
    if not isinstance(dhdl_data, list):
        dhdl_list = dhdl_data.separate_dhdl()
    else:
        dhdl_list = []
        for dhdl in dhdl_data:
            dhdl_list.extend(dhdl.separate_dhdl())

    # Convert unit
    if units is None:
        units = dhdl_list[0].attrs["energy_unit"]

    new_unit = []
    convert = get_unit_converter(units)
    for dhdl in dhdl_list:
        new_unit.append(convert(dhdl))

    dhdl_list = new_unit
    if ax is None:
        fig, ax = plt.subplots(figsize=(8, 6))

    ax.spines["bottom"].set_position("zero")
    ax.spines["top"].set_color("none")
    ax.spines["right"].set_color("none")
    ax.xaxis.set_ticks_position("bottom")
    ax.yaxis.set_ticks_position("left")

    for k, spine in ax.spines.items():
        spine.set_zorder(12.2)

    # Make level names
    if labels is None:
        lv_names2 = []
        for dhdl in dhdl_list:
            # Assume that the dhdl has only one columns
            lv_names2.append(dhdl.name.capitalize())
    else:
        if len(labels) == len(dhdl_list):
            lv_names2 = labels
        else:  # pragma: no cover
            raise ValueError(
                "Length of labels ({}) should be the same as the number of data ({})".format(
                    len(labels), len(dhdl_list)
                )
            )

    if colors is None:
        colors = ["r", "g", "#7F38EC", "#9F000F", "b", "y"]
    else:
        if len(colors) >= len(dhdl_list):
            pass
        else:  # pragma: no cover
            raise ValueError(
                "Number of colors ({}) should be larger than the number of data ({})".format(
                    len(labels), len(dhdl_list)
                )
            )

    # Get the real data out
    xs, ndx, dx = [0], 0, 0.001
    min_y, max_y = 0, 0
    for dhdl in dhdl_list:
        x = dhdl.index.values
        y = dhdl.values.ravel()

        min_y = min(y.min(), min_y)
        max_y = max(y.max(), max_y)

        for i in range(len(x) - 1):
            if i % 2 == 0:
                ax.fill_between(
                    x[i : i + 2] + ndx, 0, y[i : i + 2], color=colors[ndx], alpha=1.0
                )
            else:
                ax.fill_between(
                    x[i : i + 2] + ndx, 0, y[i : i + 2], color=colors[ndx], alpha=0.5
                )

        xlegend = [-100 * wnum for wnum in range(len(lv_names2))]
        ax.plot(
            xlegend,
            [0 * wnum for wnum in xlegend],
            ls="-",
            color=colors[ndx],
            label=lv_names2[ndx],
        )
        xs += (x + ndx).tolist()[1:]
        ndx += 1

    # Make sure the tick labels are not overcrowded.
    xs = np.array(xs)
    dl_mat = np.array([xs - i for i in xs])
    ri = range(len(xs))

    def getInd(r=ri, z=[0]):
        primo = r[0]
        min_dl = ndx * 0.02 * 2 ** (primo > 10)
        if dl_mat[primo].max() < min_dl:
            return z
        for i in r:  # pragma: no cover
            for j in range(len(xs)):
                if dl_mat[i, j] > min_dl:
                    z.append(j)
                    return getInd(ri[j:], z)

    xt = []
    for i in range(len(xs)):
        if i in getInd():
            xt.append(i)
        else:
            xt.append("")

    plt.xticks(xs[1:], xt[1:], fontsize=10)
    ax.yaxis.label.set_size(10)

    # Remove the abscissa ticks and set up the axes limits.
    for tick in ax.get_xticklines():
        tick.set_visible(False)
    ax.set_xlim(0, ndx)
    min_y *= 1.01
    max_y *= 1.01

    # Modified so that the x label won't conflict with the lambda label
    min_y -= (max_y - min_y) * 0.1

    ax.set_ylim(min_y, max_y)

    for i, j in zip(xs[1:], xt[1:]):
        ax.annotate(
            ("%.2f" % (i - 1.0 if i > 1.0 else i) if not j == "" else ""),
            xy=(i, 0),
            size=10,
            rotation=90,
            va="bottom",
            ha="center",
            color="#151B54",
        )
    if ndx > 1:
        lenticks = len(ax.get_ymajorticklabels()) - 1
        if min_y < 0:
            lenticks -= 1
        if lenticks < 5:  # pragma: no cover
            from matplotlib.ticker import AutoMinorLocator as AML

            ax.yaxis.set_minor_locator(AML())
    ax.grid(which="both", color="w", lw=0.25, axis="y", zorder=12)
    ax.set_ylabel(
        r"$\langle{\frac{\partial U}{\partial\lambda}}\rangle_{\lambda}$"
        + "({})".format(units),
        fontsize=20,
        color="#151B54",
    )
    ax.annotate(
        r"$\mathit{\lambda}$",
        xy=(0, 0),
        xytext=(0.5, -0.05),
        size=18,
        textcoords="axes fraction",
        va="top",
        ha="center",
        color="#151B54",
    )
    lege = ax.legend(prop=FP(size=14), frameon=False, loc=1)
    for l in lege.legendHandles:
        l.set_linewidth(10)
    return ax



import os
from pathlib import Path

import pandas as pd
import pytest

from alchemlyb.workflows import base


class Test_automatic_base:
    @staticmethod
    @pytest.fixture(scope="session")
    def workflow(tmp_path_factory):
        outdir = tmp_path_factory.mktemp("out") / "abc" / "def"
        workflow = base.WorkflowBase(out=str(outdir))
        workflow.run()
        return workflow

    def test_outdir(self, workflow):
        assert Path(workflow.out).is_dir()

    def test_write(self, workflow):
        """Patch the output directory to tmpdir"""
        workflow.result.to_pickle(os.path.join(workflow.out, "result.pkl"))
        assert os.path.exists(os.path.join(workflow.out, "result.pkl"))

    def test_read(self, workflow):
        assert len(workflow.u_nk_list) == 0
        assert len(workflow.dHdl_list) == 0

    def test_subsample(self, workflow):
        assert len(workflow.u_nk_sample_list) == 0
        assert len(workflow.dHdl_sample_list) == 0

    def test_estimator(self, workflow):
        assert isinstance(workflow.result, pd.DataFrame)

    def test_convergence(self, workflow):
        assert isinstance(workflow.convergence, pd.DataFrame)



import alchemlyb


def test_version():
    try:
        version = alchemlyb.__version__
    except AttributeError:
        raise AssertionError("alchemlyb does not have __version__")

    assert len(version) > 0


def test_version_get_versions():
    import alchemlyb._version

    version = alchemlyb._version.get_versions()

    assert alchemlyb.__version__ == version["version"]



"""Functions for Plotting the overlay matrix for the MBAR estimator.

To assess the quality of the MBAR estimator, the overlap matrix between
the lambda states can be computed and the more overlap is observed between
the states, the more reliable the estimate is. One way of accessing the
quality of the overlap matrix is by plotting it.

The code for producing the overlap matrix plot is taken from
`Alchemical Analysis <https://github.com/MobleyLab/alchemical-analysis>`_.

"""

import matplotlib.pyplot as plt
import numpy as np


def plot_mbar_overlap_matrix(matrix, skip_lambda_index=[], ax=None):
    """Plot the MBAR overlap matrix.

    Parameters
    ----------
    matrix : numpy.matrix
        DataFrame of the overlap matrix obtained from
        :attr:`~alchemlyb.estimators.MBAR.overlap_matrix`
    skip_lambda_index : List
        list of lambda indices to be omitted from plotting process.
        Default: ``[]``.
    ax : matplotlib.axes.Axes
        Matplotlib axes object where the plot will be drawn on. If ``ax=None``,
        a new axes will be generated.

    Returns
    -------
    matplotlib.axes.Axes
        An axes with the overlap matrix drawn.

    Note
    ----
    The code is taken and modified from
    `Alchemical Analysis <https://github.com/MobleyLab/alchemical-analysis>`_.


    .. versionadded:: 0.4.0

    """
    # Compute the size of the figure, if ax is not given.
    max_prob = matrix.max()
    size = len(matrix)
    if ax is None:
        fig, ax = plt.subplots(figsize=(size / 2, size / 2))
    ax.set_xticks([])
    ax.set_yticks([])
    ax.axis("off")
    for i in range(size):
        if i != 0:
            ax.axvline(x=i, ls="-", lw=0.5, color="k", alpha=0.25)
            ax.axhline(y=i, ls="-", lw=0.5, color="k", alpha=0.25)
        for j in range(size):
            if matrix[j, i] < 0.005:
                ii = ""
            elif matrix[j, i] > 0.995:
                ii = "1.00"
            else:
                ii = "{:.2f}".format(matrix[j, i])[1:]
            alf = matrix[j, i] / max_prob
            ax.fill_between(
                [i, i + 1],
                [size - j, size - j],
                [size - (j + 1), size - (j + 1)],
                color="k",
                alpha=alf,
            )
            ax.annotate(
                ii,
                xy=(i, j),
                xytext=(i + 0.5, size - (j + 0.5)),
                size=8,
                textcoords="data",
                va="center",
                ha="center",
                color=("k" if alf < 0.5 else "w"),
            )

    if skip_lambda_index:
        ks = [int(l) for l in skip_lambda_index]
        ks = np.delete(np.arange(size + len(ks)), ks)
    else:
        ks = range(size)
    for i in range(size):
        ax.annotate(
            ks[i],
            xy=(i + 0.5, 1),
            xytext=(i + 0.5, size + 0.5),
            size=10,
            textcoords=("data", "data"),
            va="center",
            ha="center",
            color="k",
        )
        ax.annotate(
            ks[i],
            xy=(-0.5, size - (size - 0.5)),
            xytext=(-0.5, size - (i + 0.5)),
            size=10,
            textcoords=("data", "data"),
            va="center",
            ha="center",
            color="k",
        )
    ax.annotate(
        r"$\lambda$",
        xy=(-0.5, size - (size - 0.5)),
        xytext=(-0.5, size + 0.5),
        size=10,
        textcoords=("data", "data"),
        va="center",
        ha="center",
        color="k",
    )
    ax.plot([0, size], [0, 0], "k-", lw=4.0, solid_capstyle="butt")
    ax.plot([size, size], [0, size], "k-", lw=4.0, solid_capstyle="butt")
    ax.plot([0, 0], [0, size], "k-", lw=2.0, solid_capstyle="butt")
    ax.plot([0, size], [size, size], "k-", lw=2.0, solid_capstyle="butt")

    cx = np.repeat(range(size + 1), 2)
    cy = sorted(np.repeat(range(size + 1), 2), reverse=True)
    ax.plot(cx[2:-1], cy[1:-2], "k-", lw=2.0)
    ax.plot(np.array(cx[2:-3]) + 1, cy[1:-4], "k-", lw=2.0)
    ax.plot(cx[1:-2], np.array(cy[:-3]) - 1, "k-", lw=2.0)
    ax.plot(cx[1:-4], np.array(cy[:-5]) - 2, "k-", lw=2.0)

    ax.set_xlim(-1, size)
    ax.set_ylim(0, size + 1)
    return ax



import logging
import os
from glob import glob
from os.path import join

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

from .base import WorkflowBase
from .. import concat
from ..convergence import forward_backward_convergence
from ..estimators import MBAR, BAR, TI, FEP_ESTIMATORS, TI_ESTIMATORS
from ..parsing import gmx, amber
from ..postprocessors.units import get_unit_converter
from ..preprocessing.subsampling import decorrelate_dhdl, decorrelate_u_nk
from ..visualisation import (
    plot_mbar_overlap_matrix,
    plot_ti_dhdl,
    plot_dF_state,
    plot_convergence,
)


class ABFE(WorkflowBase):
    """Workflow for absolute and relative binding free energy calculations.

    This workflow provides functionality similar to the ``alchemical-analysis.py`` script.
    It loads multiple input files from alchemical free energy calculations and computes the
    free energies between different alchemical windows using different estimators. It
    produces plots to aid in the assessment of convergence.

    Parameters
    ----------
    T : float
        Temperature in K.
    units : str
        The unit used for printing and plotting results. {'kcal/mol', 'kJ/mol',
        'kT'}. Default: 'kT'.
    software : str
        The software used for generating input (case-insensitive). {'GROMACS', 'AMBER'}.
        This option chooses the appropriate parser for the input file.
    dir : str
        Directory in which data files are stored. Default: os.path.curdir.
        The input files are searched using the pattern of
        ``dir + '/**/' + prefix + '*' + suffix``.
    prefix : str
        Prefix for datafile sets. Default: 'dhdl'.
    suffix : str
        Suffix for datafile sets. Default: 'xvg'.
    outdirectory : str
        Directory in which the output files produced by this script will be
        stored. Default: os.path.curdir.

    Attributes
    ----------
    logger : Logger
        The logging object.
    file_list : list
        The list of filenames sorted by the lambda state.


    .. versionadded:: 1.0.0
    """

    def __init__(
        self,
        T,
        units="kT",
        software="GROMACS",
        dir=os.path.curdir,
        prefix="dhdl",
        suffix="xvg",
        outdirectory=os.path.curdir,
    ):
        super().__init__(units, software, T, outdirectory)
        self.logger.info("Initialise Alchemlyb ABFE Workflow")
        self.update_units(units)
        self.logger.info(
            f"Finding files with prefix: {prefix}, suffix: "
            f"{suffix} under directory {dir} produced by "
            f"{software}"
        )
        reg_exp = dir + "/**/" + prefix + "*" + suffix
        self.file_list = list(glob(reg_exp, recursive=True))

        if len(self.file_list) == 0:
            raise ValueError(f"No file has been matched to {reg_exp}.")

        self.logger.info(f"Found {len(self.file_list)} xvg files.")
        self.logger.info("Unsorted file list: \n %s", "\n".join(self.file_list))

        if software == "GROMACS":
            self.logger.info(f"Using {software} parser to read the data.")
            self._extract_u_nk = gmx.extract_u_nk
            self._extract_dHdl = gmx.extract_dHdl
        elif software == "AMBER":
            self._extract_u_nk = amber.extract_u_nk
            self._extract_dHdl = amber.extract_dHdl
        else:
            raise NotImplementedError(f"{software} parser not found.")

    def _logger_setup(self):
        self.logger = logging.getLogger("alchemlyb.workflows.ABFE")

    def read(self, read_u_nk=True, read_dHdl=True):
        """Read the u_nk and dHdL data from the
        :attr:`~alchemlyb.workflows.ABFE.file_list`

        Parameters
        ----------
        read_u_nk : bool
            Whether to read the u_nk.
        read_dHdl : bool
            Whether to read the dHdl.

        Attributes
        ----------
        u_nk_list : list
            A list of :class:`pandas.DataFrame` of u_nk.
        dHdl_list : list
            A list of :class:`pandas.DataFrame` of dHdl.
        """
        self.u_nk_sample_list = None
        self.dHdl_sample_list = None

        u_nk_list = []
        dHdl_list = []
        for file in self.file_list:
            if read_u_nk:
                try:
                    u_nk = self._extract_u_nk(file, T=self.T)
                    self.logger.info(f"Reading {len(u_nk)} lines of u_nk from {file}")
                    u_nk_list.append(u_nk)
                except Exception as exc:
                    msg = f"Error reading u_nk from {file}."
                    self.logger.error(msg)
                    raise OSError(msg) from exc

            if read_dHdl:
                try:
                    dhdl = self._extract_dHdl(file, T=self.T)
                    self.logger.info(f"Reading {len(dhdl)} lines of dhdl from {file}")
                    dHdl_list.append(dhdl)
                except Exception as exc:
                    msg = f"Error reading dHdl from {file}."
                    self.logger.error(msg)
                    raise OSError(msg) from exc

        # Sort the files according to the state
        if read_u_nk:
            self.logger.info("Sort files according to the u_nk.")
            column_names = u_nk_list[0].columns.values.tolist()
            index_list = sorted(
                range(len(self.file_list)),
                key=lambda x: column_names.index(
                    u_nk_list[x].reset_index("time").index.values[0]
                ),
            )
        elif read_dHdl:
            self.logger.info("Sort files according to the dHdl.")
            index_list = sorted(
                range(len(self.file_list)),
                key=lambda x: dHdl_list[x].reset_index("time").index.values[0],
            )
        else:
            self.u_nk_list = []
            self.dHdl_list = []
            return

        self.file_list = [self.file_list[i] for i in index_list]
        self.logger.info("Sorted file list: \n%s", "\n".join(self.file_list))
        if read_u_nk:
            self.u_nk_list = [u_nk_list[i] for i in index_list]
        else:
            self.u_nk_list = []

        if read_dHdl:
            self.dHdl_list = [dHdl_list[i] for i in index_list]
        else:
            self.dHdl_list = []

    def run(
        self,
        skiptime=0,
        uncorr="dE",
        threshold=50,
        estimators=("MBAR", "BAR", "TI"),
        overlap="O_MBAR.pdf",
        breakdown=True,
        forwrev=None,
        *args,
        **kwargs,
    ):
        """The method for running the automatic analysis.

        Parameters
        ----------
        skiptime : float
            Discard data prior to this specified time as 'equilibration' data.
            Units are specified by the corresponding MD Engine. Default: 0.
        uncorr : str
            The observable to be used for the autocorrelation analysis; 'dE'.
        threshold : int
            Proceed with correlated samples if the number of uncorrelated samples is
            found to be less than this number. If 0 is given, the time series
            analysis will not be performed at all. Default: 50.
        estimators : str or list of str
            A list of the estimators to estimate the free energy with. Default:
            `('MBAR', 'BAR', 'TI')`.
        overlap : str
            The filename for the plot of overlap matrix. Default: 'O_MBAR.pdf'.
        breakdown : bool
            Plot the free energy differences evaluated for each pair of adjacent
            states for all methods, including the dH/dlambda curve for TI. Default:
            ``True``.
        forwrev : int
            Plot the free energy change as a function of time in both directions,
            with the specified number of points in the time plot. The number of time
            points (an integer) must be provided. Specify as ``None`` will not do
            the convergence analysis. Default: None. By default, 'MBAR'
            estimator will be used for convergence analysis, as it is
            usually the fastest converging method. If the dataset does not
            contain u_nk, please run
            meth:`~alchemlyb.workflows.ABFE.check_convergence` manually
            with estimator='TI'.

        Attributes
        ----------
        summary : Dataframe
            The summary of the free energy estimate.
        convergence : DataFrame
            The summary of the convergence results. See
            :func:`~alchemlyb.convergence.forward_backward_convergence` for
            further explanation.
        """
        use_FEP = False
        use_TI = False

        if estimators is not None:
            if isinstance(estimators, str):
                estimators = [
                    estimators,
                ]
            for estimator in estimators:
                if estimator in FEP_ESTIMATORS:
                    use_FEP = True
                elif estimator in TI_ESTIMATORS:
                    use_TI = True
                else:
                    msg = (
                        f"Estimator {estimator} is not supported. Choose one from "
                        f"{FEP_ESTIMATORS + TI_ESTIMATORS}."
                    )
                    self.logger.error(msg)
                    raise ValueError(msg)

            self.read(use_FEP, use_TI)

        if uncorr is not None:
            self.preprocess(skiptime=skiptime, uncorr=uncorr, threshold=threshold)
        if estimators is not None:
            self.estimate(estimators)
            self.generate_result()

        if overlap is not None and use_FEP:
            ax = self.plot_overlap_matrix(overlap)
            plt.close(ax.figure)

        if breakdown:
            if use_TI:
                ax = self.plot_ti_dhdl()
                plt.close(ax.figure)
            fig = self.plot_dF_state()
            plt.close(fig)
            fig = self.plot_dF_state(
                dF_state="dF_state_long.pdf", orientation="landscape"
            )
            plt.close(fig)

        if forwrev:
            ax = self.check_convergence(forwrev, estimator="MBAR", dF_t="dF_t.pdf")
            plt.close(ax.figure)

    def update_units(self, units=None):
        """Update the unit.

        Parameters
        ----------
        units : {'kcal/mol', 'kJ/mol', 'kT'}
            The unit used for printing and plotting results.

        """
        if units is not None:
            self.logger.info(f"Set unit to {units}.")
            self.units = units or None

    def preprocess(self, skiptime=0, uncorr="dE", threshold=50):
        """Preprocess the data by removing the equilibration time and
        decorrelate the date.

        Parameters
        ----------
        skiptime : float
            Discard data prior to this specified time as 'equilibration' data.
            Units are specified by the corresponding MD Engine. Default: 0.
        uncorr : str
            The observable to be used for the autocorrelation analysis; 'dE'.
        threshold : int
            Proceed with correlated samples if the number of uncorrelated
            samples is found to be less than this number. If 0 is given, the
            time series analysis will not be performed at all. Default: 50.

        Attributes
        ----------
        u_nk_sample_list : list
            The list of u_nk after decorrelation.
        dHdl_sample_list : list
            The list of dHdl after decorrelation.
        """
        self.logger.info(
            f"Start preprocessing with skiptime of {skiptime} "
            f"uncorrelation method of {uncorr} and threshold of "
            f"{threshold}"
        )
        if len(self.u_nk_list) > 0:
            self.logger.info(
                f"Processing the u_nk data set with skiptime of {skiptime}."
            )

            self.u_nk_sample_list = []
            for index, u_nk in enumerate(self.u_nk_list):
                # Find the starting frame

                u_nk = u_nk[u_nk.index.get_level_values("time") >= skiptime]
                subsample = decorrelate_u_nk(u_nk, uncorr, remove_burnin=True)

                if len(subsample) < threshold:
                    self.logger.warning(
                        f"Number of u_nk {len(subsample)} "
                        f"for state {index} is less than the "
                        f"threshold {threshold}."
                    )
                    self.logger.info(f"Take all the u_nk for state {index}.")
                    self.u_nk_sample_list.append(u_nk)
                else:
                    self.logger.info(
                        f"Take {len(subsample)} uncorrelated "
                        f"u_nk for state {index}."
                    )
                    self.u_nk_sample_list.append(subsample)
        else:
            self.logger.info("No u_nk data being subsampled")

        if len(self.dHdl_list) > 0:
            self.dHdl_sample_list = []
            for index, dHdl in enumerate(self.dHdl_list):
                dHdl = dHdl[dHdl.index.get_level_values("time") >= skiptime]
                subsample = decorrelate_dhdl(dHdl, remove_burnin=True)
                if len(subsample) < threshold:
                    self.logger.warning(
                        f"Number of dHdl {len(subsample)} for "
                        f"state {index} is less than the "
                        f"threshold {threshold}."
                    )
                    self.logger.info(f"Take all the dHdl for state {index}.")
                    self.dHdl_sample_list.append(dHdl)
                else:
                    self.logger.info(
                        f"Take {len(subsample)} uncorrelated "
                        f"dHdl for state {index}."
                    )
                    self.dHdl_sample_list.append(subsample)
        else:
            self.logger.info("No dHdl data being subsampled")

    def estimate(self, estimators=("MBAR", "BAR", "TI"), **kwargs):
        """Estimate the free energy using the selected estimator.

        Parameters
        ----------
        estimators : str or list of str
            A list of the estimators to estimate the free energy with. Default:
            ['TI', 'BAR', 'MBAR'].

        kwargs : dict
            Keyword arguments to be passed to the estimator.

        Attributes
        ----------
        estimator : dict
            The dictionary of estimators. The keys are in ['TI', 'BAR',
            'MBAR']. Note that the estimators are in their original form where
            no unit conversion has been attempted.

        """
        # Make estimators into a tuple
        if isinstance(estimators, str):
            estimators = (estimators,)

        for estimator in estimators:
            if estimator not in (FEP_ESTIMATORS + TI_ESTIMATORS):
                msg = f"Estimator {estimator} is not available in {FEP_ESTIMATORS + TI_ESTIMATORS}."
                self.logger.error(msg)
                raise ValueError(msg)

        self.logger.info(f"Start running estimator: {','.join(estimators)}.")
        self.estimator = {}
        # Use unprocessed data if preprocess is not performed.
        if "TI" in estimators:
            if self.dHdl_sample_list is not None:
                dHdl = concat(self.dHdl_sample_list)
            else:
                dHdl = concat(self.dHdl_list)
                self.logger.warning("dHdl has not been preprocessed.")
            self.logger.info(f"A total {len(dHdl)} lines of dHdl is used.")

        if "BAR" in estimators or "MBAR" in estimators:
            if self.u_nk_sample_list is not None:
                u_nk = concat(self.u_nk_sample_list)
            else:
                u_nk = concat(self.u_nk_list)
                self.logger.warning("u_nk has not been preprocessed.")
            self.logger.info(f"A total {len(u_nk)} lines of u_nk is used.")

        for estimator in estimators:
            if estimator == "MBAR":
                self.logger.info("Run MBAR estimator.")
                self.estimator[estimator] = MBAR(**kwargs).fit(u_nk)
            elif estimator == "BAR":
                self.logger.info("Run BAR estimator.")
                self.estimator[estimator] = BAR(**kwargs).fit(u_nk)
            elif estimator == "TI":
                self.logger.info("Run TI estimator.")
                self.estimator[estimator] = TI(**kwargs).fit(dHdl)

    def generate_result(self):
        """Summarise the result into a dataframe.

        Returns
        -------
        DataFrame
            The DataFrame with convergence data. ::

                                      MBAR  MBAR_Error        BAR  BAR_Error         TI  TI_Error
                States 0 -- 1     0.065967    0.001293   0.066544   0.001661   0.066663  0.001675
                       1 -- 2     0.089774    0.001398   0.089303   0.002101   0.089566  0.002144
                       2 -- 3     0.132036    0.001638   0.132687   0.002990   0.133292  0.003055
                       3 -- 4     0.116494    0.001213   0.116348   0.002691   0.116845  0.002750
                       4 -- 5     0.105251    0.000980   0.106344   0.002337   0.106603  0.002362
                       5 -- 6     0.349320    0.002781   0.343399   0.006839   0.350568  0.007393
                       6 -- 7     0.402346    0.002767   0.391368   0.006641   0.395754  0.006961
                       7 -- 8     0.322284    0.002058   0.319395   0.005333   0.321542  0.005434
                       8 -- 9     0.434999    0.002683   0.425680   0.006823   0.430251  0.007155
                       9 -- 10    0.355672    0.002219   0.350564   0.005472   0.352745  0.005591
                       10 -- 11   3.574227    0.008744   3.513595   0.018711   3.514790  0.018078
                       11 -- 12   2.896685    0.009905   2.821760   0.017844   2.823210  0.018088
                       12 -- 13   2.223769    0.011229   2.188885   0.018438   2.189784  0.018478
                       13 -- 14   1.520978    0.012526   1.493598   0.019155   1.490070  0.019288
                       14 -- 15   0.911279    0.009527   0.894878   0.015023   0.896010  0.015140
                       15 -- 16   0.892365    0.010558   0.886706   0.015260   0.884698  0.015392
                       16 -- 17   1.737971    0.025315   1.720643   0.031416   1.741028  0.030624
                       17 -- 18   1.790706    0.025560   1.788112   0.029435   1.801695  0.029244
                       18 -- 19   1.998635    0.023340   2.007404   0.027447   2.019213  0.027096
                       19 -- 20   2.263475    0.020286   2.265322   0.025023   2.282040  0.024566
                       20 -- 21   2.565680    0.016695   2.561324   0.023611   2.552977  0.023753
                       21 -- 22   1.384094    0.007553   1.385837   0.011672   1.381999  0.011991
                       22 -- 23   1.428567    0.007504   1.422689   0.012524   1.416010  0.013012
                       23 -- 24   1.440581    0.008059   1.412517   0.013125   1.408267  0.013539
                       24 -- 25   1.411329    0.009022   1.419167   0.013356   1.411446  0.013795
                       25 -- 26   1.340320    0.010167   1.360679   0.015213   1.356953  0.015260
                       26 -- 27   1.243745    0.011239   1.245873   0.015711   1.248959  0.015762
                       27 -- 28   1.128429    0.012859   1.124554   0.016999   1.121892  0.016962
                       28 -- 29   1.010313    0.016442   1.005444   0.017692   1.019747  0.017257
                Stages coul      10.215658    0.033903  10.017838   0.041839  10.017854  0.048744
                       vdw       22.547489    0.098699  22.501150   0.060092  22.542936  0.106723
                       bonded     2.374144    0.014995   2.341631   0.005507   2.363828  0.021078
                       TOTAL     35.137291    0.103580  34.860619   0.087022  34.924618  0.119206

        Attributes
        ----------
        summary : Dataframe
            The summary of the free energy estimate.
        """

        # Write estimate
        self.logger.info("Summarise the estimate into a dataframe.")
        # Make the header name
        self.logger.info("Generate the row names.")
        estimator_names = list(self.estimator.keys())
        num_states = len(self.estimator[estimator_names[0]].states_)
        data_dict = {"name": [], "state": []}
        for i in range(num_states - 1):
            data_dict["name"].append(str(i) + " -- " + str(i + 1))
            data_dict["state"].append("States")

        try:
            u_nk = self.u_nk_list[0]
            stages = u_nk.reset_index("time").index.names
            self.logger.info("use the stage name from u_nk")
        except:
            dHdl = self.dHdl_list[0]
            stages = dHdl.reset_index("time").index.names
            self.logger.info("use the stage name from dHdl")

        for stage in stages:
            data_dict["name"].append(stage.split("-")[0])
            data_dict["state"].append("Stages")
        data_dict["name"].append("TOTAL")
        data_dict["state"].append("Stages")

        col_names = []
        for estimator_name, estimator in self.estimator.items():
            self.logger.info(f"Read the results from estimator {estimator_name}")

            # Do the unit conversion
            delta_f_ = estimator.delta_f_
            d_delta_f_ = estimator.d_delta_f_
            # Write the estimator header

            col_names.append(estimator_name)
            col_names.append(estimator_name + "_Error")
            data_dict[estimator_name] = []
            data_dict[estimator_name + "_Error"] = []
            for index in range(1, num_states):
                data_dict[estimator_name].append(delta_f_.iloc[index - 1, index])
                data_dict[estimator_name + "_Error"].append(
                    d_delta_f_.iloc[index - 1, index]
                )

            self.logger.info(
                f"Generate the staged result from estimator {estimator_name}"
            )
            for index, stage in enumerate(stages):
                if len(stages) == 1:
                    start = 0
                    end = len(estimator.states_) - 1
                else:
                    # Get the start and the end of the state
                    lambda_min = min([state[index] for state in estimator.states_])
                    lambda_max = max([state[index] for state in estimator.states_])
                    if lambda_min == lambda_max:
                        # Deal with the case where a certain lambda is used but
                        # not perturbed
                        start = 0
                        end = 0
                    else:
                        states = [state[index] for state in estimator.states_]
                        start = list(reversed(states)).index(lambda_min)
                        start = num_states - start - 1
                        end = states.index(lambda_max)
                self.logger.info(f"Stage {stage} is from state {start} to state {end}.")
                # This assumes that the indexes are sorted as the
                # preprocessing should sort the index of the df.
                result = delta_f_.iloc[start, end]
                if estimator_name != "BAR":
                    error = d_delta_f_.iloc[start, end]
                else:
                    error = np.sqrt(
                        sum(
                            [
                                d_delta_f_.iloc[start, start + 1] ** 2
                                for i in range(start, end + 1)
                            ]
                        )
                    )
                data_dict[estimator_name].append(result)
                data_dict[estimator_name + "_Error"].append(error)

            # Total result
            # This assumes that the indexes are sorted as the
            # preprocessing should sort the index of the df.
            result = delta_f_.iloc[0, -1]
            if estimator_name != "BAR":
                error = d_delta_f_.iloc[0, -1]
            else:
                error = np.sqrt(
                    sum([d_delta_f_.iloc[i, i + 1] ** 2 for i in range(num_states - 1)])
                )
            data_dict[estimator_name].append(result)
            data_dict[estimator_name + "_Error"].append(error)
        summary = pd.DataFrame.from_dict(data_dict)

        summary = summary.set_index(["state", "name"])
        # Make sure that the columns are in the right order
        summary = summary[col_names]
        # Remove the name of the index column to make it prettier
        summary.index.names = [None, None]

        summary.attrs = estimator.delta_f_.attrs
        converter = get_unit_converter(self.units)
        summary = converter(summary)
        self.summary = summary
        self.logger.info(f"Write results:\n{summary.to_string()}")
        return summary

    def plot_overlap_matrix(self, overlap="O_MBAR.pdf", ax=None):
        """Plot the overlap matrix for MBAR estimator using
        :func:`~alchemlyb.visualisation.plot_mbar_overlap_matrix`.

        Parameters
        ----------
        overlap : str
            The filename for the plot of overlap matrix. Default: 'O_MBAR.pdf'
        ax : matplotlib.axes.Axes
            Matplotlib axes object where the plot will be drawn on. If
            ``ax=None``, a new axes will be generated.

        Returns
        -------
        matplotlib.axes.Axes
            An axes with the overlap matrix drawn.
        """
        self.logger.info("Plot overlap matrix.")
        if "MBAR" in self.estimator:
            ax = plot_mbar_overlap_matrix(self.estimator["MBAR"].overlap_matrix, ax=ax)
            ax.figure.savefig(join(self.out, overlap))
            self.logger.info(f"Plot overlap matrix to {self.out} under {overlap}.")
            return ax
        else:
            self.logger.warning(
                "MBAR estimator not found. " "Overlap matrix not plotted."
            )

    def plot_ti_dhdl(self, dhdl_TI="dhdl_TI.pdf", labels=None, colors=None, ax=None):
        """Plot the dHdl for TI estimator using
        :func:`~alchemlyb.visualisation.plot_ti_dhdl`.

        Parameters
        ----------
        dhdl_TI : str
            The filename for the plot of TI dHdl. Default: 'dhdl_TI.pdf'
        labels : List
            list of labels for labelling all the alchemical transformations.
        colors : List
            list of colors for plotting all the alchemical transformations.
            Default: ['r', 'g', '#7F38EC', '#9F000F', 'b', 'y']
        ax : matplotlib.axes.Axes
            Matplotlib axes object where the plot will be drawn on. If ``ax=None``,
            a new axes will be generated.

        Returns
        -------
        matplotlib.axes.Axes
            An axes with the TI dhdl drawn.
        """
        self.logger.info("Plot TI dHdl.")
        if "TI" in self.estimator:
            ax = plot_ti_dhdl(
                self.estimator["TI"],
                units=self.units,
                labels=labels,
                colors=colors,
                ax=ax,
            )
            ax.figure.savefig(join(self.out, dhdl_TI))
            self.logger.info(f"Plot TI dHdl to {dhdl_TI} under {self.out}.")
            return ax
        else:
            raise ValueError("No TI data available in estimators.")

    def plot_dF_state(
        self,
        dF_state="dF_state.pdf",
        labels=None,
        colors=None,
        orientation="portrait",
        nb=10,
    ):
        """Plot the dF states using
        :func:`~alchemlyb.visualisation.plot_dF_state`.

        Parameters
        ----------
        dF_state : str
            The filename for the plot of dF states. Default: 'dF_state.pdf'
        labels : List
            list of labels for labelling different estimators.
        colors : List
            list of colors for plotting different estimators.
        orientation : string
            The orientation of the figure. Can be `portrait` or `landscape`
        nb : int
            Maximum number of dF states in one row in the `portrait` mode

        Returns
        -------
        matplotlib.figure.Figure
            An Figure with the dF states drawn.
        """
        self.logger.info("Plot dF states.")
        fig = plot_dF_state(
            self.estimator.values(),
            labels=labels,
            colors=colors,
            units=self.units,
            orientation=orientation,
            nb=nb,
        )
        fig.savefig(join(self.out, dF_state))
        self.logger.info(f"Plot dF state to {dF_state} under {self.out}.")
        return fig

    def check_convergence(
        self, forwrev, estimator="MBAR", dF_t="dF_t.pdf", ax=None, **kwargs
    ):
        """Compute the forward and backward convergence using
        :func:`~alchemlyb.convergence.forward_backward_convergence`and
        plot with
        :func:`~alchemlyb.visualisation.plot_convergence`.

        Parameters
        ----------
        forwrev : int
            Plot the free energy change as a function of time in both
            directions, with the specified number of points in the time plot.
            The number of time points (an integer) must be provided.
        estimator : {'TI', 'BAR', 'MBAR'}
            The estimator used for convergence analysis. Default: 'MBAR'
        dF_t : str
            The filename for the plot of convergence. Default: 'dF_t.pdf'
        ax : matplotlib.axes.Axes
            Matplotlib axes object where the plot will be drawn on. If ``ax=None``,
            a new axes will be generated.
        kwargs : dict
            Keyword arguments to be passed to the estimator.

        Attributes
        ----------
        convergence : DataFrame

        Returns
        -------
        matplotlib.axes.Axes
            An axes with the convergence drawn.

        """
        self.logger.info("Start convergence analysis.")
        self.logger.info("Checking data availability.")

        if estimator in FEP_ESTIMATORS:
            if self.u_nk_sample_list is not None:
                u_nk_list = self.u_nk_sample_list
                self.logger.info("Subsampled u_nk is available.")
            else:
                if self.u_nk_list is not None:
                    u_nk_list = self.u_nk_list
                    self.logger.info(
                        "Subsampled u_nk not available, " "use original data instead."
                    )
                else:
                    msg = (
                        f"u_nk is needed for the f{estimator} estimator. "
                        f"If the dataset only has dHdl, "
                        f"run ABFE.check_convergence(estimator='TI') to "
                        f"use a TI estimator."
                    )
                    self.logger.error(msg)
                    raise ValueError(msg)
            convergence = forward_backward_convergence(
                u_nk_list, estimator=estimator, num=forwrev, **kwargs
            )
        elif estimator in TI_ESTIMATORS:
            self.logger.warning(
                "No valid FEP estimator or dataset found. " "Fallback to TI."
            )
            if self.dHdl_sample_list is not None:
                dHdl_list = self.dHdl_sample_list
                self.logger.info("Subsampled dHdl is available.")
            else:
                if self.dHdl_list is not None:
                    dHdl_list = self.dHdl_list
                    self.logger.info(
                        "Subsampled dHdl not available, " "use original data instead."
                    )
                else:
                    self.logger.error(f"dHdl is needed for the f{estimator} estimator.")
                    raise ValueError(f"dHdl is needed for the f{estimator} estimator.")
            convergence = forward_backward_convergence(
                dHdl_list, estimator=estimator, num=forwrev, **kwargs
            )
        else:
            msg = (
                f"Estimator {estimator} is not supported. Choose one from "
                f"{FEP_ESTIMATORS + TI_ESTIMATORS}."
            )
            self.logger.error(msg)
            raise ValueError(msg)

        self.convergence = get_unit_converter(self.units)(convergence)

        self.logger.info(f"Plot convergence analysis to {dF_t} under {self.out}.")

        ax = plot_convergence(self.convergence, units=self.units, ax=ax)
        ax.figure.savefig(join(self.out, dF_t))
        return ax



"""Amber parser tests.

"""
import logging

import pandas as pd
import pytest
from alchemtest.amber import load_bace_example
from alchemtest.amber import load_bace_improper
from alchemtest.amber import load_simplesolvated
from alchemtest.amber import load_testfiles
from numpy.testing import assert_allclose

from alchemlyb.parsing.amber import extract
from alchemlyb.parsing.amber import extract_dHdl
from alchemlyb.parsing.amber import extract_u_nk


##################################################################################
################ Check the parser behaviour with problematic files
##################################################################################


@pytest.fixture(name="testfiles", scope="module")
def fixture_testfiles():
    """Returns the testfiles data dictionary"""
    bunch = load_testfiles()
    return bunch["data"]


def test_file_not_found():
    with pytest.raises(FileNotFoundError):
        _ = extract("not_a_valid_file_name", T=298.0)


def test_no_dHdl_data_points(caplog, testfiles):
    """Test if we deal with a file without dHdl data points"""
    filename = testfiles["no_dHdl_data_points"][0]
    with caplog.at_level(logging.INFO):
        _ = extract(str(filename), T=298.0)
    assert "does not contain any dV/dl data" in caplog.text


def test_None_in_mbar(caplog, testfiles):
    """Test if we deal with an incorrect MBAR section"""
    filename = testfiles["none_in_mbar"][0]
    with pytest.raises(ValueError, match="strange parsing the following MBAR section"):
        with caplog.at_level(logging.ERROR):
            _ = extract(str(filename), T=298.0)
    assert "strange parsing the following MBAR section" in caplog.text


def test_unfinished_run(caplog, testfiles):
    """Test if we give a warning if we are parsing an unfinished run"""
    filename = testfiles["not_finished_run"][0]
    with caplog.at_level(logging.WARNING):
        _ = extract(str(filename), T=298.0)
    assert "is a prematurely terminated run" in caplog.text


def test_no_atomic_section(caplog, testfiles):
    """Test if raise an exception if there is no ATOMIC section"""
    filename = testfiles["no_atomic_section"][0]
    with pytest.raises(ValueError, match='no "ATOMIC" section found'):
        with caplog.at_level(logging.ERROR):
            _ = extract(str(filename), T=298.0)
    assert 'No "ATOMIC" section found' in caplog.text


def test_no_control_data(caplog, testfiles):
    """Test if we raise an exception if there is no CONTROL section"""
    filename = testfiles["no_control_data"][0]
    with pytest.raises(ValueError, match='no "CONTROL DATA" section found'):
        with caplog.at_level(logging.ERROR):
            _ = extract(str(filename), T=298.0)
    assert 'No "CONTROL DATA" section found' in caplog.text


def test_no_free_energy_info(caplog, testfiles):
    """Test if we raise an exception if there is no free energy section"""
    filename = testfiles["no_free_energy_info"][0]
    with pytest.raises(ValueError, match="no free energy section found"):
        with caplog.at_level(logging.ERROR):
            _ = extract(str(filename), T=298.0)
    assert "No free energy section found" in caplog.text


def test_no_useful_data(caplog, testfiles):
    """Test if we raise an exception if there is no data"""
    filename = testfiles["no_useful_data"][0]
    with pytest.raises(ValueError, match="does not contain any data"):
        with caplog.at_level(logging.ERROR):
            _ = extract(str(filename), T=298.0)
    assert "does not contain any data" in caplog.text


def test_no_temp0_set(caplog, testfiles):
    """Test if we raise an exception if there is no temp0 set"""
    filename = testfiles["no_temp0_set"][0]
    with pytest.raises(ValueError, match='no valid "temp0" record found'):
        with caplog.at_level(logging.ERROR):
            _ = extract(str(filename), T=298.0)
    assert 'No valid "temp0" record found' in caplog.text


def test_no_results_section(caplog, testfiles):
    """Test if we raise an exception if there is no RESULTS section"""
    filename = testfiles["no_results_section"][0]
    with pytest.raises(ValueError, match='no "RESULTS" section found'):
        with caplog.at_level(logging.ERROR):
            _ = extract(str(filename), T=298.0)
    assert 'No "RESULTS" section found' in caplog.text


def test_long_and_wrong_number_MBAR(caplog, testfiles):
    """
    Test if we have a high number of MBAR states, and also a different
    number of MBAR states than expected
    """
    filename = testfiles["high_and_wrong_number_of_mbar_windows"][0]
    with pytest.raises(ValueError, match="the number of lambda windows read"):
        with caplog.at_level(logging.ERROR):
            _ = extract_u_nk(str(filename), T=300.0)
    assert "the number of lambda windows read" in caplog.text


def test_no_starting_time(caplog, testfiles):
    """Test if raise an exception if the starting time is not read"""
    filename = testfiles["no_starting_simulation_time"][0]
    with pytest.raises(ValueError, match="No starting simulation time in file"):
        with caplog.at_level(logging.ERROR):
            _ = extract(str(filename), T=298.0)
    assert "No starting simulation time in file" in caplog.text


def test_parse_without_spaces_around_equal(testfiles):
    """
    Test if the regex is able to extract values where the are no
    spaces around the equal sign
    """
    filename = testfiles["no_spaces_around_equal"][0]
    df_dict = extract(str(filename), T=298.0)
    assert isinstance(df_dict["dHdl"], pd.DataFrame)


##################################################################################
################ Check the parser behaviour with standard single files
##################################################################################


@pytest.fixture(name="single_u_nk", scope="module")
def fixture_single_u_nk():
    """return a single file to check u_unk parsing"""
    return load_bace_example().data["complex"]["vdw"][0]


@pytest.fixture(name="single_dHdl", scope="module")
def fixture_single_dHdl():
    """return a single file to check dHdl parsing"""
    return load_simplesolvated().data["charge"][0]


def test_dHdl_time_reading(single_dHdl):
    """Test if time information is read correctly when extracting dHdl"""
    dHdl = extract_dHdl(single_dHdl, T=298.0)
    assert_allclose(dHdl.index.values[0][0], 22.0)
    assert_allclose(dHdl.index.values[-1][0], 1020.0)


def test_u_nk_time_reading(single_u_nk):
    """Test if time information is read correctly when extracting u_nk"""
    u_nk = extract_u_nk(single_u_nk, T=298.0)
    assert_allclose(u_nk.index.values[0][0], 22.0)
    assert_allclose(u_nk.index.values[-1][0], 1020.0)


def test_extract_with_both_data(single_u_nk):
    """Test that dHdl and u_nk have the correct form when
    extracted from files with the single "extract" funcion."""
    df_dict = extract(single_u_nk, T=298.0)
    assert df_dict["dHdl"].index.names == ("time", "lambdas")
    assert df_dict["dHdl"].shape == (500, 1)
    assert df_dict["u_nk"].index.names == ("time", "lambdas")


def test_extract_with_only_dhdl_data(single_dHdl):
    """Test that parsing with the extract function a file
    with just dHdl gives the correct results"""
    df_dict = extract(single_dHdl, T=298.0)
    assert df_dict["dHdl"].index.names == ("time", "lambdas")
    assert df_dict["dHdl"].shape == (500, 1)
    assert df_dict["u_nk"] is None


def test_wrong_T_should_raise_warning(single_dHdl, T=300.0):
    """
    Test if calling extract with differnt T from what's
    read from the AMBER file gives a warning
    """
    with pytest.raises(
        ValueError, match="is different from the temperature passed as parameter"
    ):
        _ = extract(single_dHdl, T=T)


###################################################################
################ Check the behaviour on proper datasets
###################################################################


@pytest.mark.parametrize(
    "filename",
    [filename for leg in load_simplesolvated()["data"].values() for filename in leg],
)
def test_dHdl(filename, names=("time", "lambdas"), shape=(500, 1)):
    """Test that dHdl has the correct form when extracted from files."""
    dHdl = extract_dHdl(filename, T=298.0)

    assert dHdl.index.names == names
    assert dHdl.shape == shape


@pytest.mark.parametrize(
    "mbar_filename",
    [
        mbar_filename
        for leg in load_bace_example()["data"]["complex"].values()
        for mbar_filename in leg
    ],
)
def test_u_nk(mbar_filename, names=("time", "lambdas")):
    """Test the u_nk has the correct form when extracted from files"""
    u_nk = extract_u_nk(mbar_filename, T=298.0)

    assert u_nk.index.names == names


@pytest.mark.parametrize(
    "improper_filename",
    [
        improper_filename
        for leg in load_bace_improper()["data"].values()
        for improper_filename in leg
    ],
)
def test_u_nk_improper(improper_filename, names=("time", "lambdas")):
    """Test the u_nk has the correct form when extracted from files"""
    try:
        u_nk = extract_u_nk(improper_filename, T=298.0)
        assert u_nk.index.names == names
    except Exception:
        assert "0.5626" in improper_filename



from functools import wraps


def _init_attrs(func):
    """Add temperature to the parsed dataframe.

    The temperature is added to the dataframe as dataframe.attrs['temperature']
    and the energy unit is initiated as dataframe.attrs['energy_unit'] = 'kT'.
    """

    @wraps(func)
    def wrapper(outfile, T, *args, **kwargs):
        dataframe = func(outfile, T, *args, **kwargs)
        if dataframe is not None:
            dataframe.attrs["temperature"] = T
            dataframe.attrs["energy_unit"] = "kT"
        return dataframe

    return wrapper


def _init_attrs_dict(func):
    """Add temperature and energy units to the parsed dataframes.

    The temperature is added to the dataframe as dataframe.attrs['temperature']
    and the energy unit is initiated as dataframe.attrs['energy_unit'] = 'kT'.
    """

    @wraps(func)
    def wrapper(outfile, T, *args, **kwargs):
        dict_with_df = func(outfile, T, *args, **kwargs)
        for k in dict_with_df.keys():
            if dict_with_df[k] is not None:
                dict_with_df[k].attrs["temperature"] = T
                dict_with_df[k].attrs["energy_unit"] = "kT"
        return dict_with_df

    return wrapper



#! /usr/bin/python
"""Setuptools-based setup script for alchemlyb.

For a basic installation just type the command::

  python setup.py install

"""

from setuptools import setup, find_packages

import versioneer

setup(
    name="alchemlyb",
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    description="the simple alchemistry library",
    author="David Dotson",
    author_email="dotsdl@gmail.com",
    maintainer="Oliver Beckstein",
    maintainer_email="orbeckst@gmail.com",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: BSD License",
        "Operating System :: POSIX",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows ",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
        "Topic :: Scientific/Engineering :: Chemistry",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    packages=find_packages("src"),
    package_dir={"": "src"},
    license="BSD",
    long_description=open("README.rst").read(),
    long_description_content_type="text/x-rst",
    python_requires=">=3.8",
    tests_require=["pytest", "alchemtest"],
    install_requires=[
        "numpy",
        "pandas>=1.4",
        "pymbar>=4",
        "scipy",
        "scikit-learn",
        "matplotlib",
    ],
)



import pandas as pd
import pytest

import alchemlyb
from alchemlyb import pass_attrs
from alchemlyb.postprocessors.units import to_kT
from alchemlyb.preprocessing import (
    dhdl2series,
    u_nk2series,
    decorrelate_u_nk,
    decorrelate_dhdl,
    slicing,
    statistical_inefficiency,
    equilibrium_detection,
)


@pytest.fixture
def dHdl(gmx_benzene_Coulomb_dHdl):
    return gmx_benzene_Coulomb_dHdl[0]


@pytest.fixture
def u_nk(gmx_benzene_Coulomb_u_nk):
    return gmx_benzene_Coulomb_u_nk[0]


def test_noT(dHdl):
    """Test no temperature error"""
    dHdl.attrs.pop("temperature", None)
    with pytest.raises(TypeError):
        to_kT(dHdl)


def test_nounit(dHdl):
    """Test no unit error"""
    dHdl.attrs.pop("energy_unit", None)
    with pytest.raises(TypeError):
        to_kT(dHdl)


def test_concat():
    """Test if different attrs could will give rise to error."""
    d = {"col1": [1, 2], "col2": [3, 4]}
    df1 = pd.DataFrame(data=d)
    df1.attrs = {1: 1}
    df2 = pd.DataFrame(data=d)
    df2.attrs = {1: 2}
    with pytest.raises(ValueError):
        alchemlyb.concat([df1, df2])


def test_concat_empty():
    """Test if empty raise the right error."""
    with pytest.raises(ValueError):
        alchemlyb.concat([])


def test_setT():
    """Test setting temperature."""
    df = pd.DataFrame(data={"col1": [1, 2]})
    df.attrs = {"temperature": 300, "energy_unit": "kT"}
    new = to_kT(df, 310)
    assert new.attrs["temperature"] == 310


class Test_Conversion:
    """Test the preprocessing module."""

    def test_kt2kt_number(self, dHdl):
        new_dhdl = to_kT(dHdl)
        assert 12.9 == pytest.approx(new_dhdl.loc[(0.0, 0.0)], 0.1)

    def test_kt2kt_unit(self, dHdl):
        new_dhdl = to_kT(dHdl)
        assert new_dhdl.attrs["energy_unit"] == "kT"

    def test_kj2kt_unit(self, dHdl):
        dHdl.attrs["energy_unit"] = "kJ/mol"
        new_dhdl = to_kT(dHdl)
        assert new_dhdl.attrs["energy_unit"] == "kT"

    def test_kj2kt_number(self, dHdl):
        dHdl.attrs["energy_unit"] = "kJ/mol"
        new_dhdl = to_kT(dHdl)
        assert 5.0 == pytest.approx(new_dhdl.loc[(0.0, 0.0)], 0.1)

    def test_kcal2kt_unit(self, dHdl):
        dHdl.attrs["energy_unit"] = "kcal/mol"
        new_dhdl = to_kT(dHdl)
        assert new_dhdl.attrs["energy_unit"] == "kT"

    def test_kcal2kt_number(self, dHdl):
        dHdl.attrs["energy_unit"] = "kcal/mol"
        new_dhdl = to_kT(dHdl)
        assert 21.0 == pytest.approx(new_dhdl.loc[(0.0, 0.0)], 0.1)

    def test_unknown2kt(self, dHdl):
        dHdl.attrs["energy_unit"] = "ddd"
        with pytest.raises(ValueError):
            to_kT(dHdl)


def test_pd_concat():
    """Test if concat will preserve the metadata.
    When this test is being made, the pd.concat will discard the attrs of
    the input dataframe. However, this should get fixed in the future.
    pandas-dev/pandas#28283
    <https://github.com/pandas-dev/pandas/issues/28283>
    """
    d = {"col1": [1, 2], "col2": [3, 4]}
    df1 = pd.DataFrame(data=d)
    df1.attrs = {1: 1}
    df2 = pd.DataFrame(data=d)
    df2.attrs = {1: 1}
    df = pd.concat([df1, df2])
    assert df.attrs == {1: 1}


def test_pass_attrs():
    d = {"col1": [1, 2], "col2": [3, 4]}
    df1 = pd.DataFrame(data=d)
    df1.attrs = {1: 1}
    df2 = pd.DataFrame(data=d)
    df2.attrs = {1: 1}

    @pass_attrs
    def concat(df1, df2):
        return pd.concat([df1, df2])

    assert concat(df1, df2).attrs == {1: 1}


def test_pd_slice():
    """Test if slicing will preserve the metadata."""
    d = {"col1": [1, 2], "col2": [3, 4]}
    df = pd.DataFrame(data=d)
    df.attrs = {1: 1}
    assert df[::2].attrs == {1: 1}


class TestRetainUnit:
    """This test tests if the functions that should retain the unit would actually
    retain the units."""

    @pytest.mark.parametrize(
        "func,fixture_in",
        [
            (dhdl2series, "dHdl"),
            (u_nk2series, "u_nk"),
            (decorrelate_u_nk, "u_nk"),
            (decorrelate_dhdl, "dHdl"),
            (slicing, "dHdl"),
            (statistical_inefficiency, "dHdl"),
            (equilibrium_detection, "dHdl"),
        ],
    )
    def test_function(self, func, fixture_in, request):
        result = func(request.getfixturevalue(fixture_in))
        assert result.attrs["energy_unit"] is not None



"""Basic building blocks for free energy workflows."""

import logging
from pathlib import Path

import pandas as pd

from .. import __version__


class WorkflowBase:
    """The base class for the Workflow.

    This is the base class for the creation of new Workflow. The
    initialisation method takes in the MD engine, unit, temperature and
    output directory. The goal of the initialisation is to check the input
    files and store them in
    :attr:`~alchemlyb.workflows.WorkflowBase.file_list`
    such that they can be read by the
    :func:`~alchemlyb.workflows.WorkflowBase.read` method.

    Parameters
    ----------
    units : string, optional
        The unit used for printing and plotting results. {'kcal/mol', 'kJ/mol',
        'kT'}. Default: 'kT'.

    software : string, optional
        The software used for generating input. {'Gromacs', 'Amber'}

    T : float, optional,
        Temperature in K. Default: 298.

    out : string, optional
        Directory in which the output files produced by this script will be
        stored. Default: './'.

    Attributes
    ----------
    file_list : list
        A list of files to be read by the parser.


    .. versionadded:: 0.7.0
    """

    def __init__(
        self, units="kT", software="Gromacs", T=298, out="./", *args, **kwargs
    ):
        self._logger_setup()
        self.logger.info(f"Alchemlyb Version: f{__version__}")
        self.logger.info(f"Set Temperature to {T} K.")
        self.T = T
        self.logger.info(f"Set Software to {software}.")
        self.software = software
        self.unit = units
        self.file_list = []
        self.out = out
        if not Path(out).is_dir():
            self.logger.info(f"Make output folder {out}.")
            Path(out).mkdir(parents=True)

    def _logger_setup(self):
        self.logger = logging.getLogger("alchemlyb.workflows.WorkflowBase")

    def run(self, *args, **kwargs):
        """Run the workflow in an automatic fashion.

        This method would execute the
        :func:`~alchemlyb.workflows.WorkflowBase.read`,
        :func:`~alchemlyb.workflows.WorkflowBase.preprocess`,
        :func:`~alchemlyb.workflows.WorkflowBase.estimate`,
        :func:`~alchemlyb.workflows.WorkflowBase.check_convergence`,
        :func:`~alchemlyb.workflows.WorkflowBase.plot`
        sequentially such that the whole analysis could be done automatically.

        This method takes in an arbitrary number of arguments and pass all
        of them to the underlying methods. The methods will be selecting the
        keywords that they would like to use.

        Running this method would generate the resulting attributes for the
        user to retrieve the results.

        Attributes
        ----------

        u_nk_list : list
            A list of :class:`pandas.DataFrame` of u_nk.
        dHdl_list : list
            A list of :class:`pandas.DataFrame` of dHdl.
        u_nk_sample_list : list
            A list of :class:`pandas.DataFrame` of the subsampled u_nk.
        dHdl_sample_list : list
            A list of :class:`pandas.DataFrame` of the subsampled dHdl.
        result : pandas.Dataframe
            The main result of the workflow.
        convergence : pandas.Dataframe
            The result of the convergence analysis.

        """
        self.read(*args, **kwargs)
        self.preprocess(*args, **kwargs)
        self.estimate(*args, **kwargs)
        self.check_convergence(*args, **kwargs)
        self.plot(*args, **kwargs)

    def read(self, *args, **kwargs):
        """The function that reads the files in `file_list` and parse them
        into u_nk and dHdl files.

        Attributes
        ----------

        u_nk_list : list
            A list of :class:`pandas.DataFrame` of u_nk.
        dHdl_list : list
            A list of :class:`pandas.DataFrame` of dHdl.

        """
        self.u_nk_list = []
        self.dHdl_list = []

    def preprocess(self, *args, **kwargs):
        """The function that subsample the u_nk and dHdl in `u_nk_list` and
        `dHdl_list`.

        Attributes
        ----------

        u_nk_sample_list : list
            A list of :class:`pandas.DataFrame` of the subsampled u_nk.
        dHdl_sample_list : list
            A list of :class:`pandas.DataFrame` of the subsampled dHdl.

        """
        self.dHdl_sample_list = []
        self.u_nk_sample_list = []

    def estimate(self, *args, **kwargs):
        """The function that runs the estimator based on `u_nk_sample_list`
        and `dHdl_sample_list`.

        Attributes
        ----------

        result : pandas.Dataframe
            The main result of the workflow.

        """
        self.result = pd.DataFrame()

    def check_convergence(self, *args, **kwargs):
        """The function for doing convergence analysis.

        Attributes
        ----------

        convergence : pandas.Dataframe
            The result of the convergence analysis.

        """
        self.convergence = pd.DataFrame()

    def plot(self, *args, **kwargs):
        """The function for producing any plots."""
        pass



import numpy as np
import pandas as pd
import pytest

from alchemlyb.convergence import forward_backward_convergence, fwdrev_cumavg_Rc, A_c
from alchemlyb.convergence.convergence import _cummean


def test_convergence_ti(gmx_benzene_Coulomb_dHdl):
    convergence = forward_backward_convergence(gmx_benzene_Coulomb_dHdl, "TI")
    assert convergence.shape == (10, 5)

    assert convergence.loc[0, "Forward"] == pytest.approx(3.07, 0.01)
    assert convergence.loc[0, "Backward"] == pytest.approx(3.11, 0.01)
    assert convergence.loc[9, "Forward"] == pytest.approx(3.09, 0.01)
    assert convergence.loc[9, "Backward"] == pytest.approx(3.09, 0.01)


@pytest.mark.parametrize("estimator", ["MBAR", "BAR"])
def test_convergence_fep(gmx_benzene_Coulomb_u_nk, estimator):
    convergence = forward_backward_convergence(gmx_benzene_Coulomb_u_nk, estimator)
    assert convergence.shape == (10, 5)
    assert convergence.loc[0, "Forward"] == pytest.approx(3.02, 0.01)
    assert convergence.loc[0, "Backward"] == pytest.approx(3.06, 0.01)
    assert convergence.loc[9, "Forward"] == pytest.approx(3.05, 0.01)
    assert convergence.loc[9, "Backward"] == pytest.approx(3.04, 0.01)


def test_convergence_wrong_estimator(gmx_benzene_Coulomb_dHdl):
    with pytest.raises(ValueError, match="is not available in"):
        forward_backward_convergence(gmx_benzene_Coulomb_dHdl, "WWW")


def test_convergence_wrong_cases(gmx_benzene_Coulomb_u_nk):
    with pytest.warns(DeprecationWarning, match="Using lower-case strings for"):
        forward_backward_convergence(gmx_benzene_Coulomb_u_nk, "mbar")


def test_convergence_method(gmx_benzene_Coulomb_u_nk):
    convergence = forward_backward_convergence(
        gmx_benzene_Coulomb_u_nk, "MBAR", num=2, method="adaptive"
    )
    assert len(convergence) == 2


def test_cummean_short():
    """Test the case where the input is shorter than the expected output"""
    value = _cummean(np.empty(10), 100)
    assert len(value) == 10


def test_cummean_long():
    """Test the case where the input is longer than the expected output"""
    value = _cummean(np.empty(20), 10)
    assert len(value) == 10


def test_cummean_long_none_integter():
    """Test the case where the input is not a integer multiple of the expected output"""
    value = _cummean(np.empty(25), 10)
    assert len(value) == 10


def test_R_c_converged():
    data = pd.Series(data=[0] * 100)
    data.attrs["temperature"] = 310
    data.attrs["energy_unit"] = "kcal/mol"
    value, running_average = fwdrev_cumavg_Rc(data)
    np.testing.assert_allclose(value, 0.0)


def test_R_c_notconverged():
    data = pd.Series(data=range(21))
    data.attrs["temperature"] = 310
    data.attrs["energy_unit"] = "kcal/mol"
    value, running_average = fwdrev_cumavg_Rc(data, tol=0.1, precision=0.05)
    np.testing.assert_allclose(value, 1.0)


def test_R_c_real():
    data = pd.Series(data=np.hstack((range(10), [4.5] * 10)))
    data.attrs["temperature"] = 310
    data.attrs["energy_unit"] = "kcal/mol"
    value, running_average = fwdrev_cumavg_Rc(data, tol=2.0)
    np.testing.assert_allclose(value, 0.35)


def test_A_c_real():
    data = pd.Series(data=np.hstack((range(10), [4.5] * 10)))
    data.attrs["temperature"] = 310
    data.attrs["energy_unit"] = "kcal/mol"
    value = A_c([data] * 2, tol=2.0)
    np.testing.assert_allclose(value, 0.65)
